import pandas as pd
from skfolio import Portfolio as BasePortfolio


class Portfolio(BasePortfolio):
    def get_next_weights(self, returns: pd.Series) -> dict[int, float]:
        """
        Given the next returns, compute the next weights of this portfolio

        Args:
            returns: The returns for the next day as a pandas series

        Returns:
            A dictionary of weights (instrument ids as keys and weights as values)
        """
        weights = self.weights_per_observation.iloc[-1, :].T
        if weights.sum() != 0:
            weights /= weights.sum()
        contribution = weights * (returns + 1.0)
        if contribution.sum() != 0:
            contribution /= contribution.sum()
        return contribution.dropna().to_dict()

    def get_estimate_net_value(self, previous_net_asset_value: float) -> float:
        if self.previous_weights is None:
            raise ValueError("No previous weights available")
        expected_returns = self.previous_weights @ self.X.iloc[-1, :].T
        return previous_net_asset_value * (1.0 + expected_returns)



from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from wbfdm.models import Instrument

from wbportfolio.models import Portfolio, PortfolioRole


class UserPortfolioRequestPermissionMixin:
    @cached_property
    def instrument(self) -> Instrument:
        return get_object_or_404(Instrument, pk=self.kwargs["instrument_id"])

    @cached_property
    def portfolio(self) -> Portfolio:
        return get_object_or_404(Portfolio, id=self.kwargs["portfolio_id"])

    @cached_property
    def instrument_or_none(self) -> Instrument | None:
        try:
            return get_object_or_404(Instrument, pk=self.kwargs["instrument_id"])
        except Exception:
            return None

    @cached_property
    def portfolio_or_none(self) -> Portfolio | None:
        try:
            return get_object_or_404(Portfolio, id=self.kwargs["portfolio_id"])
        except Exception:
            return None

    @cached_property
    def is_manager(self) -> bool:
        return PortfolioRole.is_manager(self.request.user.profile)

    @cached_property
    def is_portfolio_manager(self) -> bool:
        return PortfolioRole.is_portfolio_manager(
            self.request.user.profile, portfolio=self.portfolio_or_none, instrument=self.instrument_or_none
        )

    @cached_property
    def is_analyst(self) -> bool:
        return PortfolioRole.is_analyst(
            self.request.user.profile, portfolio=self.portfolio_or_none, instrument=self.instrument_or_none
        )

    @cached_property
    def has_portfolio_access(self) -> bool:
        return self.is_portfolio_manager or self.is_analyst



from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbfdm.serializers import InvestableUniverseRepresentationSerializer

from wbportfolio.models import Transaction
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer


class TransactionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:transaction-detail")

    class Meta:
        model = Transaction
        fields = ("id", "transaction_date", "total_value", "_detail")


class TransactionModelSerializer(wb_serializers.ModelSerializer):
    external_id = wb_serializers.CharField(required=False, read_only=True)
    value_date = wb_serializers.DateField(required=False, read_only=True)
    total_value_usd = wb_serializers.FloatField(default=0, read_only=True, label="Total Value ($)")
    total_value_gross_usd = wb_serializers.FloatField(default=0, read_only=True, label="Total Value Gross ($)")
    transaction_underlying_type = wb_serializers.CharField(read_only=True)
    transaction_url_type = wb_serializers.SerializerMethodField()
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _underlying_instrument = InvestableUniverseRepresentationSerializer(source="underlying_instrument")
    _currency = CurrencyRepresentationSerializer(source="currency")

    total_value = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True)
    total_value_fx_portfolio = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True)
    total_value_gross = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True)
    total_value_gross_fx_portfolio = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True)

    def get_transaction_url_type(self, obj):
        return obj.transaction_type.lower()

    # def get_transaction_underlying_type(self, obj):
    #     try:
    #         casted_transaction = obj.get_casted_transaction()
    #         return casted_transaction.Type[casted_transaction.transaction_subtype].label
    #     except Exception as e:
    #         return ""
    class Meta:
        model = Transaction
        decorators = {
            "total_value": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_gross": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_usd": wb_serializers.decorator(decorator_type="text", position="left", value="{{$}}"),
            "total_value_gross_usd": wb_serializers.decorator(decorator_type="text", position="left", value="{{$}}"),
            "total_value_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
            "total_value_gross_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
            # "total_value_fx_portfolio": wb_serializers.decorator(decorator_type="text", position="left", value="{{_portfolio.currency_symbol}}}"),
            # "total_value_gross_fx_portfolio": wb_serializers.decorator(decorator_type="text", position="left", value="{{_portfolio.currency_symbol}}"),
        }
        fields = (
            "id",
            "transaction_type",
            "transaction_url_type",
            "transaction_underlying_type",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "comment",
            "total_value_usd",
            "total_value_gross_usd",
        )



import random

import factory

from wbportfolio.models import DividendTransaction

from .transactions import TransactionFactory


class DividendTransactionsFactory(TransactionFactory):
    class Meta:
        model = DividendTransaction

    shares = factory.LazyAttribute(lambda o: random.randint(10, 10000))
    price = factory.LazyAttribute(lambda o: random.randint(10, 10000))



from contextlib import suppress
from datetime import date
from decimal import Decimal

from django.conf import settings
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Company, CustomerStatus
from wbcore.models import WBModel
from wbcore.utils.importlib import import_from_dotted_path

from wbportfolio.models import Claim, Product


def get_total_assets_under_management(val_date: date) -> Decimal:
    return sum([product.get_total_aum_usd(val_date) for product in Product.active_objects.all()])


def get_lost_client_customer_status():
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbportfolio__lost_client_customer_status"]


def get_returning_client_customer_status():
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbportfolio__returning_client_customer_status"]


def get_tpm_customer_status():
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbportfolio__tpm_customer_status"]


def get_client_customer_status():
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbportfolio__client_customer_status"]


class Updater:
    def __init__(self, val_date: date):
        self.val_date = val_date
        self.total_assets_under_management = get_total_assets_under_management(val_date)

    def update_company_data(self, company):
        # save company portfolio data
        company_portfolio_data = CompanyPortfolioData.objects.get_or_create(company=company)[0]
        if (
            invested_assets_under_management_usd := company_portfolio_data.get_assets_under_management_usd(
                self.val_date
            )
        ) is not None:
            company_portfolio_data.invested_assets_under_management_usd = invested_assets_under_management_usd
        if (potential := company_portfolio_data.get_potential(self.val_date)) is not None:
            company_portfolio_data.potential = potential
        company_portfolio_data.save()

        # update the company object itself
        if (tier := company_portfolio_data.get_tiering(self.total_assets_under_management)) is not None:
            company.tier = tier
        company.customer_status = company_portfolio_data.get_customer_status()
        company.save()

    # def update_all_companies(self, val_date: date):
    #     for company in tqdm(qs, total=qs.count()):
    #         with suppress(CompanyPortfolioData.DoesNotExist):
    #             company_portfolio = CompanyPortfolioData.objects.get(company=company)
    #             company_portfolio.update_data(date.today())


class CompanyPortfolioData(models.Model):
    class InvestmentDiscretion(models.TextChoices):
        FULLY_DISCRETIONAIRY = "FULLY_DISCRETIONAIRY", "Fully Discretionairy"
        MOSTLY_DISCRETIONAIRY = "MOSTLY_DISCRETIONAIRY", "Mostly Discretionairy"
        MIXED = "MIXED", "Mixed"
        MOSTLY_ADVISORY = "MOSTLY_ADVISORY", "Mostly Advisory"
        FULLY_ADVISORY = "FULLY_ADVISORY", "Fully Advisory"

    potential_help_text = """
        The potential reflects how much potential a company (regardless whether client/propective) has. The formula to calculate the potential is:

        AUM * Asset Allocation Percent * Asset Allocation Max Investment - Invested AUM.
    """

    company = models.OneToOneField(
        to="directory.Company", related_name="portfolio_data", on_delete=models.CASCADE, unique=True
    )

    assets_under_management_currency = models.ForeignKey(
        to="currency.Currency",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="AUM Currency",
    )

    assets_under_management = models.DecimalField(
        max_digits=17,
        decimal_places=2,
        null=True,
        blank=True,
        verbose_name="AUM",
        help_text="The Assets under Management (AUM) that is managed by this company or this person's primary employer.",
    )
    invested_assets_under_management_usd = models.DecimalField(
        max_digits=17,
        decimal_places=2,
        null=True,
        blank=True,
        help_text="The invested Assets under Management (AUM).",
        verbose_name="Invested AUM ($)",
    )

    investment_discretion = models.CharField(
        max_length=21,
        choices=InvestmentDiscretion.choices,
        default=InvestmentDiscretion.MIXED,
        help_text="What discretion this company or this person's primary employer has to invest its assets.",
        verbose_name="Investment Discretion",
    )

    potential = models.DecimalField(
        decimal_places=2, max_digits=19, null=True, blank=True, help_text=potential_help_text
    )
    potential_currency = models.ForeignKey(
        to="currency.Currency",
        related_name="wbportfolio_potential_currencies",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
    )

    def get_assets_under_management_usd(self, val_date: date) -> Decimal:
        return Claim.objects.filter(status=Claim.Status.APPROVED).filter_for_customer(
            self.company
        ).annotate_asset_under_management_for_date(val_date).aggregate(
            invested_aum_usd=models.Sum("asset_under_management_usd")
        )["invested_aum_usd"] or Decimal(0)

    def _get_default_potential(self, val_date: date) -> Decimal:
        with suppress(CurrencyFXRates.DoesNotExist):
            fx = CurrencyFXRates.objects.get(currency=self.assets_under_management_currency, date=val_date).value

            aum_usd = self.assets_under_management / fx
            aum_potential = Decimal(0)
            for asset_allocation in self.company.asset_allocations.all():
                aum_potential += aum_usd * asset_allocation.percent * asset_allocation.max_investment
            invested_aum = self.invested_assets_under_management_usd or Decimal(0.0)

            return aum_potential - invested_aum

    def get_potential(self, val_date: date) -> Decimal:
        if module_path := getattr(settings, "PORTFOLIO_COMPANY_DATA_POTENTIAL_METHOD", None):
            with suppress(ModuleNotFoundError):
                return import_from_dotted_path(module_path)(self, val_date)
        return self._get_default_potential(val_date)

    def _get_default_tiering(self, total_asset_under_management: Decimal) -> Company.Tiering:
        if self.company.customer_status and self.company.customer_status in [
            get_client_customer_status(),
            get_tpm_customer_status(),
        ]:
            invested_aum = self.invested_assets_under_management_usd or Decimal(0)
            match invested_aum / total_asset_under_management:
                case share if share >= 0.1:  # noqa: F821
                    return Company.Tiering.ONE
                case share if share >= 0.05:  # noqa: F821
                    return Company.Tiering.TWO
                case share if share >= 0.02:  # noqa: F821
                    return Company.Tiering.THREE
                case share if share >= 0.01:  # noqa: F821
                    return Company.Tiering.FOUR
                case _:
                    return Company.Tiering.FIVE

        elif self.assets_under_management and self.assets_under_management_currency:
            fx = self.assets_under_management_currency.fx_rates.latest("date").value

            match self.assets_under_management / fx:
                case aum if aum >= 10_000_000_000:  # noqa: F821
                    return Company.Tiering.ONE
                case aum if aum >= 5_000_000_000:  # noqa: F821
                    return Company.Tiering.TWO
                case aum if aum >= 1_000_000_000:  # noqa: F821
                    return Company.Tiering.THREE
                case aum if aum >= 500_000_000:  # noqa: F821
                    return Company.Tiering.FOUR
                case _:
                    return Company.Tiering.FIVE

        return None

    def get_tiering(self, total_asset_under_management: Decimal) -> Decimal:
        if module_path := getattr(settings, "PORTFOLIO_COMPANY_DATA_TIERING_METHOD", None):
            with suppress(ModuleNotFoundError):
                return import_from_dotted_path(module_path)(self, total_asset_under_management)
        return self._get_default_tiering(total_asset_under_management)

    def get_customer_status(self) -> CustomerStatus:
        if aum := self.invested_assets_under_management_usd:
            if aum > 0 and self.company.customer_status == get_lost_client_customer_status():
                return get_returning_client_customer_status()

            if aum > 0 and self.company.customer_status not in [
                get_tpm_customer_status(),
                get_returning_client_customer_status(),
            ]:
                return get_client_customer_status()

        if (
            not self.invested_assets_under_management_usd
            and self.company.customer_status == get_client_customer_status()
        ):
            return get_lost_client_customer_status()

        return self.company.customer_status

    def __str__(self) -> str:
        return f"{self.company}"

    class Meta:
        verbose_name = "Company Portfolio Data"
        verbose_name_plural = "Company Portfolio Data"


@receiver(post_save, sender="directory.Company")
def create_company_portfolio_data(sender, instance, created, **kwargs):
    CompanyPortfolioData.objects.get_or_create(company=instance)


class AssetAllocationType(WBModel):
    name = models.CharField(max_length=255)
    default_max_investment = models.DecimalField(
        decimal_places=4,
        max_digits=5,
        default=0.1,
        help_text="The default percentage this allocation is counted towards the potential.",
    )

    def __str__(self) -> str:
        return f"{self.name}"

    class Meta:
        verbose_name = "Asset Allocation Type"
        verbose_name_plural = "Asset Allocation Types"

    @classmethod
    def get_endpoint_basename(cls):
        return "company_portfolio:assetallocationtype"

    @classmethod
    def get_representation_endpoint(cls):
        return "company_portfolio:assetallocationtyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class AssetAllocation(models.Model):
    company = models.ForeignKey(to="directory.Company", related_name="asset_allocations", on_delete=models.CASCADE)
    asset_type = models.ForeignKey(
        to="company_portfolio.AssetAllocationType", related_name="asset_allocations", on_delete=models.PROTECT
    )
    percent = models.DecimalField(decimal_places=4, max_digits=5)
    max_investment = models.DecimalField(
        decimal_places=4,
        max_digits=5,
        null=True,
        blank=True,
        help_text="The percentage this allocation is counted towards the potential. Defaults to the default provided in the asset type.",
    )
    comment = models.TextField(default="")

    def save(self, *args, **kwargs):
        # If max investment is none, we are using the default one given by the asset_type
        if self.max_investment is None:
            self.max_investment = self.asset_type.default_max_investment

        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.company}: {self.percent:.2%} {self.asset_type}"

    class Meta:
        verbose_name = "Asset Allocation"
        verbose_name_plural = "Asset Allocations"

    @classmethod
    def get_endpoint_basename(cls):
        return "company_portfolio:assetallocation"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{company}}: {{percent}} {{asset_type}}"


class GeographicFocus(models.Model):
    company = models.ForeignKey(to="directory.Company", related_name="geographic_focuses", on_delete=models.CASCADE)
    country = models.ForeignKey(to="geography.Geography", on_delete=models.PROTECT, verbose_name="Location")
    percent = models.DecimalField(decimal_places=4, max_digits=5)
    comment = models.TextField(default="")

    def __str__(self) -> str:
        return f"{self.company}: {self.percent:.2%} {self.country}"

    class Meta:
        verbose_name = "Geographic Focus"
        verbose_name_plural = "Geographic Focuses"

    @classmethod
    def get_endpoint_basename(cls):
        return "company_portfolio:geographicfocus"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{company}}: {{percent}} {{country}}"



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from .viewsets import (
    AssetAllocationModelViewSet,
    AssetAllocationTypeModelViewSet,
    AssetAllocationTypeRepresentationViewSet,
    GeographicFocusModelViewSet,
)

router = WBCoreRouter()

router.register(
    r"assetallocationtyperepresentation",
    AssetAllocationTypeRepresentationViewSet,
    basename="assetallocationtyperepresentation",
)
router.register(r"assetallocationtype", AssetAllocationTypeModelViewSet, basename="assetallocationtype")
router.register(r"assetallocation", AssetAllocationModelViewSet, basename="assetallocation")
router.register(r"geographicfocus", GeographicFocusModelViewSet, basename="geographicfocus")

company_router = WBCoreRouter()
company_router.register(r"companyassetallocation", AssetAllocationModelViewSet, basename="companyassetallocation")
company_router.register(r"companygeographicfocus", GeographicFocusModelViewSet, basename="companygeographicfocus")

urlpatterns = [
    path("", include(router.urls)),
    path("crmcompany/<int:company_id>/", include(company_router.urls)),
]



from decimal import Decimal

import pytest
from django.test.client import RequestFactory
from faker import Faker
from pandas.tseries.offsets import BDay, BMonthEnd, BYearEnd
from rest_framework.reverse import reverse
from rest_framework.test import force_authenticate

from wbportfolio.factories import InstrumentPriceFactory, ProductFactory
from wbportfolio.viewsets.product_performance import PerformanceComparisonPandasView

fake = Faker()


@pytest.mark.django_db
class TestPerformanceComparisonPandasView:
    @classmethod
    def populated_product_and_get_prices(cls):
        product = ProductFactory.create()
        last_date = (fake.date_object() - BDay(0)).date()
        date_t1 = (last_date - BDay(1)).date()  # t-1 date
        date_m1 = (last_date - BMonthEnd(1)).date()  # last month latest date
        date_y1 = (last_date - BYearEnd(1)).date()  # last year latest date
        date_inception = (last_date - BYearEnd(2)).date()  # some random date to represent inception

        last_price = InstrumentPriceFactory.create(instrument=product, date=last_date, calculated=False)
        t1 = InstrumentPriceFactory.create(instrument=product, date=date_t1, calculated=False)
        InstrumentPriceFactory.create(instrument=product, date=date_t1, calculated=True)  # Create noise

        m1 = InstrumentPriceFactory.create(instrument=product, date=date_m1, calculated=False)
        InstrumentPriceFactory.create(instrument=product, date=date_m1, calculated=True)  # Create noise
        InstrumentPriceFactory.create(
            instrument=product, date=(fake.date_between(date_m1, date_t1) - BDay(0)).date(), calculated=False
        )  # Create noise

        y1 = InstrumentPriceFactory.create(instrument=product, date=date_y1, calculated=False)
        InstrumentPriceFactory.create(instrument=product, date=date_y1, calculated=True)  # Create noise
        InstrumentPriceFactory.create(
            instrument=product, date=(fake.date_between(date_y1, date_m1) - BDay(0)).date(), calculated=False
        )  # Create noise

        inception = InstrumentPriceFactory.create(instrument=product, date=date_inception, calculated=False)
        InstrumentPriceFactory.create(instrument=product, date=date_inception, calculated=True)  # Create noise
        InstrumentPriceFactory.create(
            instrument=product, date=(fake.date_between(date_inception, date_y1) - BDay(0)).date(), calculated=False
        )  # Create noise

        product.last_valuation_date = last_date
        product.inception_date = date_inception
        product.save()
        last_price.refresh_from_db()
        m1.refresh_from_db()

        return last_price.net_value, t1.net_value, m1.net_value, y1.net_value, inception.net_value

    def test_simple_performance(self, superuser):
        """
        Basic test to check that the given performance are correct
        """
        last_price, t1, m1, y1, inception = self.populated_product_and_get_prices()

        url = reverse("wbportfolio:productperformancecomparison-list")
        request = RequestFactory().get(url)
        force_authenticate(request, user=superuser)
        view = PerformanceComparisonPandasView.as_view({"get": "list"})
        response = view(request)
        res = response.data["results"][0]
        assert pytest.approx(Decimal(res["instrument_last_valuation_price"]), rel=Decimal(1e-4)) == last_price
        assert pytest.approx(Decimal(res["perf_last_day"]), rel=Decimal(1e-4)) == last_price / t1 - Decimal(1)
        assert pytest.approx(Decimal(res["perf_month_to_date"]), rel=Decimal(1e-4)) == last_price / m1 - Decimal(1)
        assert pytest.approx(Decimal(res["perf_year_to_date"]), rel=Decimal(1e-4)) == last_price / y1 - Decimal(1)
        assert pytest.approx(Decimal(res["perf_inception"]), rel=Decimal(1e-4)) == last_price / inception - Decimal(1)



from contextlib import suppress
from datetime import date
from decimal import Decimal
from typing import Optional

import numpy as np
import pandas as pd
from django.contrib import admin
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from wbcore.contrib.currency.models import CurrencyFXRates
from wbfdm.models.instruments import Instrument, InstrumentPrice


class PMSInstrument(Instrument):
    class Meta:
        proxy = True

    @property
    def total_aum_usd(self):
        if last_nav := self.get_latest_valid_price():
            return self.get_total_aum_usd(last_nav.date)
        return Decimal(0)

    @property
    def net_value_key(self):
        return "net_value"

    @property
    @admin.display(description="Invested")
    def _is_invested(self) -> bool:
        """
        Return True if the associated asset portfolio is invested
        """
        return getattr(self, "is_invested", self.portfolio and self.portfolio.is_invested_at_date(date.today()))

    def get_total_aum_usd(self, val_date: date) -> Decimal:
        """
        Get portfolio value at date into usd
        """
        try:
            last_price = self.prices.filter(date__lte=val_date).latest("date")
            try:
                fx_rate = CurrencyFXRates.objects.get(currency=self.currency, date=last_price.date).value
            except CurrencyFXRates.DoesNotExist:
                fx_rate = Decimal(1.0)
            return self.total_shares(last_price.date) * last_price.net_value * fx_rate
        except InstrumentPrice.DoesNotExist:
            return Decimal(0)

    def total_shares(self, val_date):
        from wbportfolio.models.transactions.trades import Trade

        trades = Trade.valid_customer_trade_objects.filter(
            underlying_instrument=self,
        )
        if val_date:
            trades = trades.filter(transaction_date__lt=val_date)
        return trades.aggregate(s=models.Sum("shares"))["s"] or Decimal(0)

    def nominal_value(self, val_date):
        return self.total_shares(val_date) * self.share_price

    def get_latest_price(self, val_date: date) -> InstrumentPrice | None:
        try:
            return InstrumentPrice.objects.filter_only_valid_prices().get(instrument=self, date=val_date)
        except InstrumentPrice.DoesNotExist:
            if not self.inception_date or not self.prices.filter(date__lte=val_date).exists():
                return InstrumentPrice.objects.get_or_create(
                    instrument=self, date=val_date, defaults={"calculated": False, "net_value": self.issue_price}
                )[0]

    def get_latest_valid_price(self, val_date: Optional[date] = None) -> models.Model:
        qs = self.valuations.exclude(net_value=0)
        if val_date and qs.filter(date__lte=val_date).exists():
            qs = qs.filter(date__lte=val_date)
        return qs.latest("date") if qs.exists() else None

    def get_earliest_valid_price(self, val_date: Optional[date] = None) -> models.Model:
        qs = self.valuations.exclude(net_value=0)
        if val_date:
            qs = qs.filter(date__gte=val_date)
        return qs.earliest("date") if qs.exists() else None

    def get_price_range(self, val_date: Optional[date] = None) -> dict:
        prices = self.valuations.order_by("net_value")
        if val_date:
            prices = prices.filter(date__lte=val_date)
        if prices.exists():
            low = prices.first()
            high = prices.last()
            return {
                "high": {"price": float(high.net_value), "date": high.date},
                "low": {"price": float(low.net_value), "date": low.date},
            }
        return {}

    def get_cumulative_shares(self, from_date: date, to_date: date) -> pd.Series:
        from wbportfolio.models.transactions import Trade

        df = pd.DataFrame(
            Trade.valid_customer_trade_objects.filter(underlying_instrument=self).values("transaction_date", "shares"),
            columns=["transaction_date", "shares"],
        )
        df["shares"] = df["shares"].astype(float)
        df["transaction_date"] = df["transaction_date"] + pd.tseries.offsets.BDay(1)
        df["transaction_date"] = pd.to_datetime(df["transaction_date"])
        ts = pd.bdate_range(from_date, to_date, freq="B")
        try:
            shares = df.groupby("transaction_date").sum(numeric_only=True)["shares"].sort_index().fillna(0).cumsum()
            return shares.reindex(ts, method="ffill").fillna(0)
        except KeyError:
            return pd.Series(0, index=ts).rename("shares")

    def update_outstanding_shares(self, clear: bool = False):
        if (from_date := self.inception_date) and (to_date := self.last_price_date):
            update_objs = []
            cumshares = self.get_cumulative_shares(from_date, to_date)
            df = pd.concat([cumshares, (cumshares - cumshares.shift(1)).rename("volume")], axis=1)
            df["volume"] = df["volume"].astype(float)
            df = df.fillna(0)  # missing shares or volume is replaced by 0
            df["volume_50d"] = df.volume.rolling(50).mean()
            df["volume_200d"] = df.volume.rolling(50).mean()
            df = df.replace([np.inf, -np.inf, np.nan], None)

            for val_date, row in df.to_dict("index").items():
                with suppress(InstrumentPrice.DoesNotExist):
                    shares = row["shares"]
                    volume = row["volume"]
                    price = InstrumentPrice.objects.get(instrument=self, date=val_date, calculated=True)
                    if (price.outstanding_shares != row["shares"]) or (price.volume != volume) or clear:
                        price.outstanding_shares = shares
                        price.outstanding_shares_consolidated = shares
                        price.volume = volume
                        price.volume_50d = row["volume_50d"]
                        price.volume_200d = row["volume_200d"]
                        update_objs.append(price)
            InstrumentPrice.objects.bulk_update(
                update_objs,
                ["outstanding_shares", "outstanding_shares_consolidated", "volume", "volume_50d", "volume_200d"],
            )


class PMSInstrumentAbstractModel(PMSInstrument):
    net_asset_value_computation_method_path = models.CharField(
        null=True,
        blank=True,
        default="wbportfolio.models.portfolio.default_estimate_net_value",
        verbose_name="NAV Computation Method",
    )
    risk_scale = models.IntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(7)],
        default=4,
        verbose_name="Risk Scale",
    )

    class Meta:
        abstract = True



# Generated by Django 5.0.10 on 2025-01-28 10:17

import django.db.models.expressions
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0071_alter_trade_options_alter_trade_order'),
    ]

    operations = [
        migrations.AddField(
            model_name='trade',
            name='diff_shares',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('shares'), '-', models.F('claimed_shares')), output_field=models.DecimalField(decimal_places=4, max_digits=15)),
        ),
    ]



from django.db import models

from wbportfolio.import_export.handlers.dividend import DividendImportHandler

from .transactions import ShareMixin, Transaction


class DividendTransaction(Transaction, ShareMixin, models.Model):
    import_export_handler_class = DividendImportHandler
    retrocession = models.FloatField(default=1)

    def save(self, *args, **kwargs):
        if (
            self.shares is not None
            and self.price is not None
            and self.retrocession is not None
            and self.total_value is None
        ):
            self.total_value = self.shares * self.price * self.retrocession

        if self.price is not None and self.price_gross is None:
            self.price_gross = self.price

        if (
            self.price_gross is not None
            and self.retrocession is not None
            and self.shares is not None
            and self.total_value_gross is None
        ):
            self.total_value_gross = self.shares * self.price_gross * self.retrocession

        super().save(*args, **kwargs)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AdjustmentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            return reverse("wbportfolio:instrument-adjustment-list", args=[instrument_id], request=self.request)
        return reverse("wbportfolio:adjustment-list", args=[], request=self.request)

    def get_instance_endpoint(self, **kwargs):
        if self.view.kwargs.get("pk", None):
            return None
        return self.get_endpoint(**kwargs)

    def get_delete_endpoint(self, **kwargs):
        return None



from decimal import Decimal

import pytest
from pandas._libs.tslibs.offsets import BDay
from wbfdm.models import InstrumentPrice

from wbportfolio.factories import PortfolioFactory, TradeFactory, TradeProposalFactory
from wbportfolio.models import PortfolioPortfolioThroughModel, Trade, TradeProposal


@pytest.mark.django_db
class TestEquallyWeightedRebalancing:
    @pytest.fixture()
    def model(self, portfolio, weekday):
        from wbportfolio.rebalancing.models import EquallyWeightedRebalancing

        return EquallyWeightedRebalancing(portfolio, (weekday + BDay(1)).date(), weekday)

    def test_is_valid(self, portfolio, weekday, model, asset_position_factory):
        assert not model.is_valid()
        asset_position_factory.create(portfolio=model.portfolio, date=model.last_effective_date)
        assert model.is_valid()

    def test_get_target_portfolio(self, portfolio, weekday, model, asset_position_factory):
        a1 = asset_position_factory(weighting=0.7, portfolio=portfolio, date=model.last_effective_date)
        a2 = asset_position_factory(weighting=0.3, portfolio=portfolio, date=model.last_effective_date)
        target_portfolio = model.get_target_portfolio()
        target_positions = target_portfolio.positions_map
        assert target_positions[a1.underlying_instrument.id].weighting == Decimal(0.5)
        assert target_positions[a2.underlying_instrument.id].weighting == Decimal(0.5)


@pytest.mark.django_db
class TestModelPortfolioRebalancing:
    @pytest.fixture()
    def model(self, portfolio, weekday):
        from wbportfolio.rebalancing.models import ModelPortfolioRebalancing

        return ModelPortfolioRebalancing(portfolio, (weekday + BDay(1)).date(), weekday)

    def test_is_valid(self, portfolio, weekday, model, asset_position_factory):
        assert not model.is_valid()
        asset_position_factory.create(portfolio=model.portfolio, date=model.last_effective_date)
        assert not model.is_valid()
        model_portfolio = PortfolioFactory.create()
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=model.portfolio,
            dependency_portfolio=model_portfolio,
            type=PortfolioPortfolioThroughModel.Type.MODEL,
        )
        asset_position_factory.create(portfolio=model.model_portfolio, date=model.last_effective_date)
        assert model.is_valid()

    def test_get_target_portfolio(self, portfolio, weekday, model, asset_position_factory):
        model_portfolio = PortfolioFactory.create()
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=model.portfolio,
            dependency_portfolio=model_portfolio,
            type=PortfolioPortfolioThroughModel.Type.MODEL,
        )
        asset_position_factory(portfolio=portfolio, date=model.last_effective_date)  # noise
        asset_position_factory(portfolio=portfolio, date=model.last_effective_date)  # noise
        a1 = asset_position_factory(weighting=0.8, portfolio=portfolio.model_portfolio, date=model.last_effective_date)
        a2 = asset_position_factory(weighting=0.2, portfolio=portfolio.model_portfolio, date=model.last_effective_date)
        target_portfolio = model.get_target_portfolio()
        target_positions = target_portfolio.positions_map
        assert target_positions[a1.underlying_instrument.id].weighting == Decimal("0.800000")
        assert target_positions[a2.underlying_instrument.id].weighting == Decimal("0.200000")


@pytest.mark.django_db
class TestCompositeRebalancing:
    @pytest.fixture()
    def model(self, portfolio, weekday):
        from wbportfolio.rebalancing.models import CompositeRebalancing

        return CompositeRebalancing(portfolio, (weekday + BDay(1)).date(), weekday)

    def test_is_valid(self, portfolio, weekday, model, asset_position_factory):
        assert not model.is_valid()

        trade_proposal = TradeProposalFactory.create(
            portfolio=model.portfolio, trade_date=model.trade_date, status=TradeProposal.Status.APPROVED
        )
        TradeFactory.create(
            portfolio=model.portfolio,
            transaction_date=model.trade_date,
            transaction_subtype=Trade.Type.BUY,
            trade_proposal=trade_proposal,
            weighting=0.7,
            status=Trade.Status.EXECUTED,
        )
        TradeFactory.create(
            portfolio=model.portfolio,
            transaction_date=model.trade_date,
            transaction_subtype=Trade.Type.BUY,
            trade_proposal=trade_proposal,
            weighting=0.3,
            status=Trade.Status.EXECUTED,
        )
        assert model.is_valid()

    def test_get_target_portfolio(self, portfolio, weekday, model, asset_position_factory):
        trade_proposal = TradeProposalFactory.create(
            portfolio=model.portfolio, trade_date=model.trade_date, status=TradeProposal.Status.APPROVED
        )
        asset_position_factory(portfolio=portfolio, date=model.last_effective_date)  # noise
        asset_position_factory(portfolio=portfolio, date=model.last_effective_date)  # noise
        t1 = TradeFactory.create(
            portfolio=model.portfolio,
            transaction_date=model.trade_date,
            transaction_subtype=Trade.Type.BUY,
            trade_proposal=trade_proposal,
            weighting=0.8,
            status=Trade.Status.EXECUTED,
        )
        t2 = TradeFactory.create(
            portfolio=model.portfolio,
            transaction_date=model.trade_date,
            transaction_subtype=Trade.Type.BUY,
            trade_proposal=trade_proposal,
            weighting=0.2,
            status=Trade.Status.EXECUTED,
        )
        target_portfolio = model.get_target_portfolio()
        target_positions = target_portfolio.positions_map
        assert target_positions[t1.underlying_instrument.id].weighting == Decimal("0.800000")
        assert target_positions[t2.underlying_instrument.id].weighting == Decimal("0.200000")


@pytest.mark.django_db
class TestMarketCapitalizationRebalancing:
    @pytest.fixture()
    def model(self, portfolio, weekday, instrument_factory, instrument_price_factory):
        from wbportfolio.rebalancing.models import MarketCapitalizationRebalancing

        last_effective_date = (weekday - BDay(1)).date()

        i1 = instrument_factory()
        i2 = instrument_factory()
        instrument_price_factory.create(instrument=i1, date=last_effective_date)
        instrument_price_factory.create(instrument=i1, date=weekday)
        instrument_price_factory.create(instrument=i2, date=last_effective_date)  # weekday is ffill
        return MarketCapitalizationRebalancing(portfolio, weekday, last_effective_date, instrument_ids=[i1.id, i2.id])

    def test_is_valid(self, portfolio, weekday, model, instrument_factory, instrument_price_factory):
        assert not model.is_valid()
        i2 = model.market_cap_df.columns[1]
        model.market_cap_df.loc[weekday, i2] = 1000  # some value
        assert model.is_valid()

    def test_get_target_portfolio(self, portfolio, weekday, model, asset_position_factory):
        i1 = model.market_cap_df.columns[0]
        i2 = model.market_cap_df.columns[1]
        mkt12 = InstrumentPrice.objects.get(instrument_id=i1, date=model.trade_date).market_capitalization
        mkt21 = InstrumentPrice.objects.get(instrument_id=i2, date=model.last_effective_date).market_capitalization

        target_portfolio = model.get_target_portfolio()
        target_positions = target_portfolio.positions_map
        assert target_positions[i1].weighting == mkt12 / (mkt12 + mkt21)
        assert target_positions[i2].weighting == mkt21 / (mkt12 + mkt21)



import operator
from functools import reduce

import numpy as np
import pandas as pd

from wbportfolio.models import Trade

FIELD_MAP = {
    "Transaction_Date": "transaction_date",
    "Trade_Date": "book_date",
    "Value_Date": "value_date",
    "Price": "price",
    "Quantity": "total_value",
    "Currency": "currency__key",
    "Portfolio_Identifier": "underlying_instrument__identifier",
    "External_Reference": "external_identifier2",
    "Booking_Comment": "comment",
}

CUSTOMER_TRADE_ORDER_TYPES = [
    "stex_issue_primary_cnt",
    "stex_buy_secondary_cnt",
    "stex_sell_secondary",
    "sectrx2_rdmpt",
    "stex_sell_secondary_cnt",
]


def parse(import_source):
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.loc[
        df["Order_Type"].isin(CUSTOMER_TRADE_ORDER_TYPES),
        :,
    ]
    data = []
    if not df.empty:
        df = df.rename(columns=FIELD_MAP)
        df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
        df["shares"] = df["total_value"] / df["price"]
        df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["book_date"] = pd.to_datetime(df["book_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["value_date"] = pd.to_datetime(df["value_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["transaction_subtype"] = df.shares.apply(
            lambda x: Trade.Type.REDEMPTION if x < 0 else Trade.Type.SUBSCRIPTION
        )
        certificate_keys = import_source.source.import_parameters.get(
            "certificate_keys", ["Index", "Strategic Certificate"]
        )
        df = df.dropna(subset=["underlying_instrument__identifier"])
        df = df[reduce(operator.or_, [df["comment"].str.contains(key, case=False) for key in certificate_keys])]
        df = df.replace([np.inf, -np.inf, np.nan], None)
        df["bank"] = "N/A"
        data = df.to_dict("records")
    return {"data": data}



from wbcore import filters
from wbcore.utils.date import financial_year_to_date
from wbcrm.models.accounts import Account

from wbportfolio.models.products import Product
from wbportfolio.models.transactions.claim import Claim


class AssetsAndNetNewMoneyProgressionFilterSet(filters.FilterSet):
    period = filters.DateRangeFilter(
        label="Period",
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
        default=financial_year_to_date,
    )

    product = filters.ModelChoiceFilter(
        label="Product",
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key=Product.get_representation_label_key(),
        queryset=Product.objects.all(),
        lookup_expr="exact",
        field_name="product",
        method=lambda queryset, label, value: queryset,
    )

    account = filters.ModelChoiceFilter(
        label="Account",
        endpoint=Account.get_representation_endpoint(),
        value_key=Account.get_representation_value_key(),
        label_key=Account.get_representation_label_key(),
        queryset=Account.objects.all(),
        lookup_expr="exact",
        field_name="account",
        filter_params={"level": 0},
        method=lambda queryset, label, value: queryset,
    )

    class Meta:
        model = Claim
        fields = {}



from django.contrib import admin

from wbportfolio.admin.transactions import TransactionModelAdmin
from wbportfolio.models import Trade, TradeProposal


@admin.register(Trade)
class TradeAdmin(TransactionModelAdmin):
    search_fields = [*TransactionModelAdmin.search_fields, "bank"]
    list_filter = (*TransactionModelAdmin.list_filter, "pending")
    list_display = (
        "transaction_subtype",
        "status",
        "shares",
        "price",
        "pending",
        "marked_for_deletion",
        "exclude_from_history",
        *TransactionModelAdmin.list_display[1:],
    )
    readonly_fields = [
        "_effective_weight",
        "_target_weight",
        "_effective_shares",
        "_target_shares",
    ]
    fieldsets = (
        ("Transaction Information", TransactionModelAdmin.fieldsets[0][1]),
        (
            "Trade Information",
            {
                "fields": (
                    ("transaction_subtype", "status", "pending", "marked_for_deletion", "exclude_from_history"),
                    ("price", "price_gross"),
                    ("register", "custodian", "bank", "external_identifier2"),
                    ("_effective_weight", "_target_weight", "weighting"),
                    ("_effective_shares", "_target_shares", "shares"),
                )
            },
        ),
    )
    autocomplete_fields = [*TransactionModelAdmin.autocomplete_fields, "register", "custodian"]


@admin.register(TradeProposal)
class TradeProposalAdmin(admin.ModelAdmin):
    search_fields = ["portfolio__name", "comment"]

    list_display = ("portfolio", "rebalancing_model", "trade_date", "status")
    autocomplete_fields = ["portfolio", "rebalancing_model"]



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_portfolio_manager

EQUITYCASHPOSITION_MENUITEM = MenuItem(
    label="Portfolio Cash Position",
    endpoint="wbportfolio:productcashposition-list",
    permission=ItemPermission(permissions=["wbfdm.view_instrument"], method=is_portfolio_manager),
)



from wbportfolio.models import DividendTransaction

from .transactions import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)


class DividendRepresentationSerializer(TransactionRepresentationSerializer):
    class Meta:
        model = DividendTransaction
        fields = TransactionRepresentationSerializer.Meta.fields


class DividendModelSerializer(TransactionModelSerializer):
    class Meta:
        model = DividendTransaction
        fields = TransactionModelSerializer.Meta.fields



from wbportfolio.models import Product


def get_product(identifier):
    return Product.objects.filter(identifier=identifier).first()


def get_portfolio_id(identifier):
    if (product := get_product(identifier)) and (portfolio := product.primary_portfolio):
        return portfolio.id


def get_perf_fee_isin(source):
    default_perf_fee_isin = "CH0040602242"
    if source:
        return source.import_parameters.get("performance_fees_instrument_isin", default_perf_fee_isin)
    return default_perf_fee_isin



import pandas as pd

from wbportfolio.pms.analytics.portfolio import Portfolio


def test_get_next_weights():
    w0 = 0.3
    w1 = 0.5
    w2 = 0.2
    r0 = 0.1
    r1 = 0.05
    r2 = -0.23
    weights = [w0, w1, w2]
    returns = [r0, r1, r2]
    portfolio = Portfolio(X=pd.DataFrame([returns]), weights=pd.Series(weights))
    next_weights = portfolio.get_next_weights(pd.Series(returns))

    assert next_weights[0] == w0 * (r0 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1))
    assert next_weights[1] == w1 * (r1 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1))
    assert next_weights[2] == w2 * (r2 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1))


def test_get_estimate_net_value():
    w0 = 0.3
    w1 = 0.5
    w2 = 0.2
    r0 = 0.1
    r1 = 0.05
    r2 = -0.23
    weights = [w0, w1, w2]
    returns = [r0, r1, r2]
    portfolio = Portfolio(X=pd.DataFrame([returns]), weights=pd.Series(weights), previous_weights=pd.Series(weights))
    current_price = 100
    net_asset_value = portfolio.get_estimate_net_value(current_price)
    return net_asset_value == current_price * (1.0 + w0 * r0 + w1 * r1 + w2 * r2)



import random

import factory
from faker import Faker
from wbfdm.factories import InstrumentFactory, InstrumentTypeFactory

from wbportfolio.models import Index

fake = Faker()


class IndexFactory(InstrumentFactory):
    risk_scale = factory.LazyAttribute(lambda o: random.randint(1, 7))

    instrument_type = factory.LazyAttribute(lambda o: InstrumentTypeFactory.create(name="Index", key="index"))

    class Meta:
        model = Index



from wbportfolio.pms.typing import Portfolio
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Model Portfolio Rebalancing")
class ModelPortfolioRebalancing(AbstractRebalancingModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def model_portfolio_rel(self):
        return self.portfolio.dependency_through.filter(type="MODEL").first()

    @property
    def model_portfolio(self):
        if model_portfolio_rel := self.model_portfolio_rel:
            return model_portfolio_rel.dependency_portfolio

    def is_valid(self) -> bool:
        return (
            self.model_portfolio.assets.filter(date=self.last_effective_date).exists()
            if self.model_portfolio
            else False
        )

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        assets = self.model_portfolio.get_positions(self.last_effective_date)

        for asset in assets:
            asset.date = self.trade_date
            asset.asset_valuation_date = self.trade_date
            positions.append(asset._build_dto())
        return Portfolio(positions=tuple(positions))



# Import CRM relevant data

from .adjustments import Adjustment
from .asset import AssetPosition, AssetPositionGroupBy
from .custodians import Custodian
from .products import Product, FeeProductPercentage
from .indexes import Index
from .product_groups import ProductGroup, ProductGroupRepresentant
from .portfolio_relationship import (
    PortfolioInstrumentPreferredClassificationThroughModel,
    InstrumentPortfolioThroughModel,
    PortfolioBankAccountThroughModel,
)
from .portfolio import Portfolio, PortfolioPortfolioThroughModel
from .portfolio_cash_targets import PortfolioCashTarget
from .portfolio_cash_flow import DailyPortfolioCashFlow
from .portfolio_swing_pricings import PortfolioSwingPricing
from .registers import Register
from .roles import PortfolioRole
from .transactions import *
from .reconciliations import AccountReconciliation, AccountReconciliationLine



# Generated by Django 4.1.9 on 2023-06-27 07:21

import django.db.models.deletion
import django_fsm
import wbcore.utils.strings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0005_account_accountrole_accountroletype_and_more"),
        ("wbportfolio", "0045_alter_instrument_financial_instrument"),
    ]

    operations = [
        migrations.AddField(
            model_name="product",
            name="default_account",
            field=models.ForeignKey(
                blank=True,
                help_text="If a default Account is set, then all newly created tradesare automatically matched to this sub account.",
                limit_choices_to=models.Q(("is_terminal_account", True)),
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="default_product_accounts",
                to="wbcrm.account",
                verbose_name="Default Account",
            ),
        ),
        migrations.CreateModel(
            name="Claim",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("APPROVED", "Approved"),
                            ("WITHDRAWN", "Withdrawn"),
                            ("DRAFT", "Draft"),
                            ("AUTO_MATCHED", "Auto-Matched"),
                        ],
                        default="DRAFT",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                ("date", models.DateField(verbose_name="Date")),
                ("bank", models.CharField(blank=True, max_length=255, verbose_name="Bank")),
                (
                    "reference",
                    models.CharField(blank=True, max_length=255, null=True, verbose_name="Reference"),
                ),
                (
                    "shares",
                    models.DecimalField(
                        blank=True,
                        decimal_places=4,
                        help_text="The amount of shares purchased / sold.",
                        max_digits=15,
                        null=True,
                        verbose_name="Shares",
                    ),
                ),
                (
                    "nominal_amount",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        help_text="The nominal amount purchased / sold. Either shares or nominal_amount has to be provided.",
                        max_digits=15,
                        null=True,
                        verbose_name="Nominal Amount",
                    ),
                ),
                (
                    "external_id",
                    models.CharField(
                        blank=True,
                        help_text="An External ID to reference a claim.",
                        max_length=255,
                        null=True,
                        verbose_name="External Identifier",
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to=models.Q(("is_terminal_account", True)),
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="claims",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "claimant",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="claimed",
                        to="directory.entry",
                        verbose_name="Claimant",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="claims_created",
                        to="directory.entry",
                        verbose_name="Creator",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="claims",
                        to="wbportfolio.product",
                        verbose_name="Product",
                    ),
                ),
                (
                    "trade",
                    models.ForeignKey(
                        blank=True,
                        help_text="Please select a Product first. The customer-trade the claim is consolidated against. The customer-trade and the claim don't necessarily have to have the same date, number of shares, etc.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="claims",
                        to="wbportfolio.trade",
                        verbose_name="Trade",
                    ),
                ),
            ],
            options={
                "verbose_name": "Claim",
                "verbose_name_plural": "Claims",
            },
            bases=(wbcore.utils.strings.ReferenceIDMixin, models.Model),
        ),
        migrations.AlterField(
            model_name="portfoliorole",
            name="role_type",
            field=models.CharField(
                choices=[
                    ("MANAGER", "Manager"),
                    ("RISK_MANAGER", "Risk Manager"),
                    ("PORTFOLIO_MANAGER", "Portfolio Manager"),
                    ("ANALYST", "Analyst"),
                ],
                max_length=24,
                verbose_name="Type",
            ),
        ),
    ]



import pandas as pd
from pandas.tseries.offsets import BDay

from .utils import file_name_parse

FIELD_MAP = {
    "Date": "transaction_date",
    "Manag. Fees Natixis": "ISSUER",
    "Manag. Fees Client": "MANAGEMENT",
    "Perf fees amount": "PERFORMANCE"
    # "Currency": "currency"
}


def parse(import_source):
    # Parse the Parts of the filename into the different parts
    parts = file_name_parse(import_source.file.name)
    # Get the valuation date and investment from the parts list
    parts["valuation_date"]
    product = parts["product"]

    df = pd.read_csv(import_source.file, encoding="utf-8", delimiter=";")
    df = df.rename(columns=FIELD_MAP)
    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
    df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True)

    # Switch the weeekend day fees to the next monday
    df["transaction_date"] = df["transaction_date"].apply(lambda x: x + BDay(1) if x.weekday() in [5, 6] else x)

    # Ensure float columns are number
    for col in ["MANAGEMENT", "ISSUER", "PERFORMANCE"]:
        df[col] = df[col].astype("str").str.replace(" ", "").astype("float")

    # Groupby and sum similar fees (e.g. Monday)
    df = df.groupby("transaction_date").sum().reset_index()
    df = pd.melt(
        df,
        id_vars=["transaction_date"],
        value_vars=["MANAGEMENT", "ISSUER", "PERFORMANCE"],
        var_name="transaction_subtype",
        value_name="total_value",
    )
    df = df[df["total_value"] != 0]

    df["portfolio"] = product.primary_portfolio.id
    df["linked_product"] = product.id
    df["transaction_date"] = df["transaction_date"].dt.strftime("%Y-%m-%d")
    df["calculated"] = False
    df["total_value_gross"] = df["total_value"]
    df["total_value_fx_portfolio"] = df["total_value"]
    df["total_value_gross_fx_portfolio"] = df["total_value"]

    return {"data": df.to_dict("records")}



from decimal import Decimal

import pandas as pd
from django.db.models import Q, QuerySet
from wbfdm.enums import MarketData
from wbfdm.models import (
    Classification,
    Exchange,
    Instrument,
    InstrumentClassificationThroughModel,
    InstrumentListThroughModel,
)

from wbportfolio.pms.typing import Portfolio, Position
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Market Capitalization Rebalancing")
class MarketCapitalizationRebalancing(AbstractRebalancingModel):
    TARGET_CURRENCY: str = "USD"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        instruments = self._get_instruments(**kwargs)
        self.market_cap_df = pd.DataFrame(
            instruments.dl.market_data(
                values=[MarketData.MARKET_CAPITALIZATION],
                from_date=self.last_effective_date,
                to_date=self.trade_date,
                target_currency=self.TARGET_CURRENCY,
            )
        )
        self.exchange_df = pd.DataFrame(
            instruments.values_list("id", "exchange"), columns=["id", "exchange"]
        ).set_index("id")
        try:
            self.market_cap_df = self.market_cap_df[
                ["valuation_date", "market_capitalization", "instrument_id"]
            ].pivot_table(index="valuation_date", columns="instrument_id", values="market_capitalization")
            # self.market_cap_df = df.iloc[-1, :].transpose()
        except (IndexError, KeyError):
            self.market_cap_df = pd.DataFrame()

    def _get_instruments(
        self,
        classification_ids: list[int] | None = None,
        instrument_ids: list[int] | None = None,
        instrument_list_id: int | None = None,
    ) -> QuerySet[Instrument]:
        """
        Use the provided kwargs to return a list of instruments as universe.
        - If classifications are given, we returns all the instrument linked to these classifications
        - Or directly from a static list of instrument ids
        - fallback to the last effective portfolio underlying instruments list
        """
        if not instrument_ids:
            instrument_ids = []
        if classification_ids:
            classifications = set()
            for classification in Classification.objects.filter(id__in=classification_ids):
                for children in classification.get_descendants(include_self=True):
                    classifications.add(children)
            instrument_ids.extend(
                list(
                    InstrumentClassificationThroughModel.objects.filter(
                        classification__in=classifications
                    ).values_list("id", flat=True)
                )
            )
        if instrument_list_id:
            instrument_ids.extend(
                list(
                    InstrumentListThroughModel.objects.filter(instrument_list_id=instrument_list_id).values_list(
                        "instrument", flat=True
                    )
                )
            )

        if not instrument_ids:
            instrument_ids = list(
                self.portfolio.assets.filter(date=self.last_effective_date).values_list(
                    "underlying_instrument", flat=True
                )
            )

        return Instrument.objects.filter(id__in=instrument_ids).filter(
            (Q(delisted_date__isnull=True) | Q(delisted_date__gt=self.trade_date))
        )

    def is_valid(self) -> bool:
        if not self.market_cap_df.empty:
            trade_date_mktp_cap = (
                self.market_cap_df.loc[self.trade_date, :].transpose().rename("market_capitalization")
            )
            df = pd.concat(
                [trade_date_mktp_cap, self.exchange_df], axis=1
            )  # if we are missing any market cap for not-delisted instrument, we consider the rebalancing not valid
            df = df.groupby("exchange", dropna=False)["market_capitalization"].any()
            missing_exchanges = Exchange.objects.filter(id__in=df[~df].index.to_list())
            if missing_exchanges.exists():
                setattr(
                    self,
                    "_validation_errors",
                    f"Couldn't find any market capitalization for exchanges {', '.join([str(e) for e in missing_exchanges])}",
                )
            return df.all()
        return False

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        market_cap_df = self.market_cap_df.ffill().loc[self.trade_date, :].transpose()
        total_market_cap = market_cap_df.sum()

        for underlying_instrument, market_cap in market_cap_df.to_dict().items():
            positions.append(
                Position(
                    underlying_instrument=underlying_instrument,
                    weighting=Decimal(market_cap / total_market_cap),
                    date=self.trade_date,
                )
            )
        return Portfolio(positions=tuple(positions))



from .transactions import TransactionModelAdmin  # isort:skip
from .claim import ClaimModelAdmin
from .dividends import DividendAdmin
from .fees import FeesAdmin
from .trades import TradeAdmin
from .rebalancing import RebalancingModelAdmin, RebalancerAdmin



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_portfolio_manager

PortfolioMenuItem = MenuItem(
    label="Portfolios",
    endpoint="wbportfolio:portfolio-list",
    permission=ItemPermission(method=is_portfolio_manager, permissions=["wbportfolio.view_portfolio"]),
)


ModelPortfolioMenuItem = MenuItem(
    label="Managed Portfolios",
    endpoint="wbportfolio:portfolio-list",
    endpoint_get_parameters={"is_manageable": True},
    permission=ItemPermission(method=is_portfolio_manager, permissions=["wbportfolio.view_portfolio"]),
)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class FeeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbportfolio:fees-list", args=[], request=self.request)


class FeesPortfolioEndpointConfig(FeeEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-fees-list", args=[self.view.kwargs["portfolio_id"]], request=self.request
        )


class FeesAggregatedPortfolioPandasEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-feesaggregated-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )



import pytest

from wbportfolio.serializers import ClaimModelSerializer


@pytest.mark.django_db
class TestClaimModelSerializer:
    def test_serialize(self, claim):
        claim.trade_type = True
        claim.quantity = 1000
        serializer = ClaimModelSerializer(claim)
        assert serializer.data

    def test_validate(self):
        data = {
            "trade_type": True,
            "quantity": 1000,
            "as_shares": True,
            "date": "2024-01-01",
        }
        serializer = ClaimModelSerializer(data=data)
        assert serializer.is_valid()



from datetime import date, timedelta

from django.db import models
from django.db.models import Q
from django.dispatch import receiver
from wbcore.contrib.directory.signals import deactivate_profile
from wbcore.signals import pre_merge
from wbfdm.models import Instrument


class PortfolioRole(models.Model):
    default_error_messages = {
        "manager": "If role type is Manager or Risk Manager, no {model} can be selected.",
        "product_index": "Not both Product and Index can be selected.",
        "start_end": "If a start and end date are selected, then the end date has be after the start date.",
    }

    class RoleType(models.TextChoices):
        MANAGER = "MANAGER", "Manager"
        RISK_MANAGER = "RISK_MANAGER", "Risk Manager"
        PORTFOLIO_MANAGER = "PORTFOLIO_MANAGER", "Portfolio Manager"
        ANALYST = "ANALYST", "Analyst"

    role_type = models.CharField(max_length=24, choices=RoleType.choices, verbose_name="Type")
    person = models.ForeignKey(
        "directory.Person",
        related_name="portfolio_roles",
        on_delete=models.CASCADE,
        verbose_name="Person",
    )

    start = models.DateField(null=True, blank=True, verbose_name="Start")
    end = models.DateField(null=True, blank=True, verbose_name="End")

    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        related_name="portfolio_roles",
        null=True,
        blank=True,
        limit_choices_to=models.Q(children__isnull=True),
        on_delete=models.CASCADE,
        verbose_name="Instrument",
    )

    weighting = models.FloatField(default=1, verbose_name="Weight")

    class Meta:
        verbose_name = "Portfolio Role"
        verbose_name_plural = "Portfolio Roles"

    def __str__(self):
        return f"{self.role_type} {self.person.computed_str}"

    def save(self, *args, **kwargs):
        assert (
            self.role_type in [self.RoleType.MANAGER, self.RoleType.RISK_MANAGER] and not self.instrument
        ) or self.role_type in [
            self.RoleType.PORTFOLIO_MANAGER,
            self.RoleType.ANALYST,
        ], self.default_error_messages[
            "manager"
        ].format(
            model="instrument"
        )

        assert (self.start and self.end and self.start < self.end) or (
            not self.start or not self.end
        ), self.default_error_messages["start_end"]

        super().save(*args, **kwargs)

    @classmethod
    def is_manager(cls, person):
        """Determines whether a person can access manager things (or is a superuser)

        Arguments:
            person {crm.Person} -- The person to be checked

        Returns:
            bool -- Can manage or not
        """

        today = date.today()

        manager = cls.objects.filter(
            Q(person=person)
            & Q(role_type__in=[cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            & (Q(start__isnull=True) | Q(start__lte=today))
            & (Q(end__isnull=True) | Q(end__gte=today))
        )

        return person.user_account.is_superuser or manager.exists()

    @classmethod
    def is_portfolio_manager(cls, person, instrument=None, portfolio=None, include_superuser=True):
        """Determines whether a person can access portfolio manager things (or is a superuser/manager)

        Arguments:
            person {crm.Person} -- The person to be checked

        Returns:
            bool -- Can manage or not
        """
        today = date.today()
        manager = cls.objects.filter(
            (
                Q(person=person)
                & Q(role_type__in=[cls.RoleType.PORTFOLIO_MANAGER, cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            )
            & (Q(start__isnull=True) | Q(start__lte=today))
            & (Q(end__isnull=True) | Q(end__gte=today))
        )

        if instrument:
            manager = manager.filter(
                Q(instrument=instrument)
                | Q(instrument__isnull=True)
                | Q(role_type__in=[cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            )
        elif portfolio:
            manager = manager.filter(
                Q(instrument__in=portfolio.instruments.all())
                | Q(instrument__isnull=True)
                | Q(role_type__in=[cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            )

        if include_superuser:
            return person.user_account.is_superuser or manager.exists()
        return manager.exists()

    @classmethod
    def is_analyst(cls, person, instrument=None, portfolio=None, include_superuser=True):
        """Determines whether a person can access analyst things (or is a superuser/manager/portfolio manager)

        Arguments:
            person {crm.Person} -- The person to be checked

        Returns:
            bool -- Can manage or not
        """
        today = date.today()
        manager = cls.objects.filter(
            Q(person=person)
            & (Q(start__isnull=True) | Q(start__lte=today))
            & (Q(end__isnull=True) | Q(end__gte=today))
        )

        if instrument:
            manager = manager.filter(
                Q(instrument=instrument)
                | Q(instrument__isnull=True)
                | Q(role_type__in=[cls.RoleType.PORTFOLIO_MANAGER, cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            )
        elif portfolio:
            manager = manager.filter(
                Q(instrument__in=portfolio.instruments.all())
                | Q(instrument__isnull=True)
                | Q(role_type__in=[cls.RoleType.PORTFOLIO_MANAGER, cls.RoleType.MANAGER, cls.RoleType.RISK_MANAGER])
            )
        if include_superuser:
            return person.user_account.is_superuser or manager.exists()
        return manager.exists()

    @classmethod
    def portfolio_managers(cls):
        today = date.today()
        return (
            cls.objects.filter(
                Q(role_type=cls.RoleType.PORTFOLIO_MANAGER)
                & (Q(start__isnull=True) | Q(start__lte=today))
                & (Q(end__isnull=True) | Q(end__gte=today))
            )
            .values("person")
            .distinct("person")
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfoliorole"


@receiver(deactivate_profile)
def handle_user_deactivation(sender, instance, substitute_profile=None, **kwargs):
    deactivation_date = date.today() - timedelta(days=1)
    for role in PortfolioRole.objects.filter(person=instance, end__isnull=True):
        role.end = deactivation_date
        role.save()
        if (
            substitute_profile
            and not PortfolioRole.objects.filter(
                person=substitute_profile, end__isnull=True, instrument=role.instrument
            ).exists()
        ):
            PortfolioRole.objects.create(
                instrument=role.instrument,
                person=substitute_profile,
                role_type=role.role_type,
                start=deactivation_date,
                weighting=role.weighting,
            )


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: Instrument, main_object: Instrument, **kwargs):
    """
    Simply reassign the portfolio roles linked to the merged instrument to the main instrument
    """
    merged_object.portfolio_roles.update(instrument=main_object)



from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class TradeProposalButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                key="replay",
                icon=WBIcon.SYNCHRONIZE.icon,
                label="Replay Trades",
                description_fields="""
                <p>Replay Trades. It will recompute all assets positions until next trade proposal day (or today otherwise) </p>
                """,
                action_label="Replay Trade",
                title="Replay Trade",
            )
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



import logging
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import TypeVar

from celery import shared_task
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django_fsm import FSMField, transition
from pandas._libs.tslibs.offsets import BDay
from wbcompliance.models.risk_management.mixins import RiskCheckMixin
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbfdm.models.instruments.instruments import Instrument

from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.trading import TradingService
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import TradeBatch as TradeBatchDTO

from .. import AssetPosition
from .trades import Trade

logger = logging.getLogger("pms")

SelfTradeProposal = TypeVar("SelfTradeProposal", bound="TradeProposal")


class TradeProposal(RiskCheckMixin, WBModel):
    trade_date = models.DateField(verbose_name="Trading Date")

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        SUBMIT = "SUBMIT", "Submit"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"
        FAILED = "FAILED", "Failed"

    comment = models.TextField(default="", verbose_name="Trade Comment", blank=True)
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    rebalancing_model = models.ForeignKey(
        "wbportfolio.RebalancingModel",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="trade_proposals",
        verbose_name="Rebalancing Model",
        help_text="Rebalancing Model that generates the target portfolio",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="trade_proposals", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    creator = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="trade_proposals",
        on_delete=models.PROTECT,
        verbose_name="Owner",
    )

    class Meta:
        verbose_name = "Trade Proposal"
        verbose_name_plural = "Trade Proposals"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "trade_date"],
                name="unique_trade_proposal",
            ),
        ]

    def save(self, *args, **kwargs):
        if not self.trade_date and self.portfolio.assets.exists():
            self.trade_date = (self.portfolio.assets.latest("date").date + BDay(1)).date()
        if not self.rebalancing_model and (rebalancer := getattr(self.portfolio, "automatic_rebalancer", None)):
            self.rebalancing_model = rebalancer.rebalancing_model
        super().save(*args, **kwargs)

    def _get_checked_object_field_name(self) -> str:
        """
        Mandatory function from the Riskcheck mixin that returns the field (aka portfolio), representing the object to check the rules against.
        """
        return "portfolio"

    @cached_property
    def validated_trading_service(self) -> TradingService:
        """
        This property holds the validated trading services and cache it.This property expect to be set only if is_valid return True
        """
        return TradingService(
            self.trade_date,
            effective_portfolio=self.portfolio._build_dto(self.trade_date),
            trades_batch=self._build_dto(),
        )

    @cached_property
    def last_effective_date(self) -> date:
        try:
            return self.portfolio.assets.filter(date__lt=self.trade_date).latest("date").date
        except AssetPosition.DoesNotExist:
            return (self.trade_date - BDay(1)).date()

    @property
    def previous_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__lt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.latest("trade_date")
        return None

    @property
    def next_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__gt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.earliest("trade_date")
        return None

    @property
    def base_assets(self) -> dict[int, Decimal]:
        """
        Return a dictionary representation (instrument_id: target weight) of this trade proposal
        Returns:
            A dictionary representation

        """
        return {
            v["underlying_instrument"]: v["target_weight"]
            for v in self.trades.all()
            .annotate_base_info()
            .filter(status=Trade.Status.EXECUTED)
            .values("underlying_instrument", "target_weight")
        }

    def __str__(self) -> str:
        return f"{self.portfolio.name}: {self.trade_date} ({self.status})"

    def _build_dto(self) -> TradeBatchDTO:
        """
        Data Transfer Object
        Returns:
            DTO trade object
        """
        return (
            TradeBatchDTO(tuple([trade._build_dto() for trade in self.trades.all()])) if self.trades.exists() else None
        )

    # Start tools methods
    def clone(self, **kwargs) -> SelfTradeProposal:
        """
        Method to clone self as a new trade proposal. It will automatically shift the trade date if a proposal already exists
        Args:
            **kwargs: The keyword arguments
        Returns:
            The cloned trade proposal
        """
        trade_date = kwargs.get("trade_date", self.trade_date)

        # Find the next valid trade date
        while TradeProposal.objects.filter(portfolio=self.portfolio, trade_date=trade_date).exists():
            trade_date += timedelta(days=1)

        trade_proposal_clone = TradeProposal.objects.create(
            trade_date=trade_date,
            comment=kwargs.get("comment", self.comment),
            status=TradeProposal.Status.DRAFT,
            rebalancing_model=self.rebalancing_model,
            portfolio=self.portfolio,
            creator=self.creator,
        )
        return trade_proposal_clone

    def normalize_trades(self):
        """
        Call the trading service with the existing trades and normalize them in order to obtain a total sum target weight of 100%
        The existing trade will be modified directly with the given normalization factor
        """
        service = TradingService(self.trade_date, trades_batch=self._build_dto())
        service.normalize()
        leftovers_trades = self.trades.all()
        for _, trade in service.trades_batch.trades_map.items():
            with suppress(Trade.DoesNotExist):
                trade = Trade.objects.get(id=trade.id)
                trade.weighting = trade.delta_weight
                trade.shares = self.estimate_shares(trade)
                trade.save()
                leftovers_trades = leftovers_trades.exclude(id=trade.id)
        leftovers_trades.delete()

    def _get_target_portfolio(self, **kwargs) -> PortfolioDTO:
        if self.rebalancing_model:
            params = {}
            if rebalancer := getattr(self.portfolio, "automatic_rebalancer", None):
                params.update(rebalancer.parameters)
            params.update(kwargs)
            return self.rebalancing_model.get_target_portfolio(
                self.portfolio, self.trade_date, self.last_effective_date, **params
            )
        # Return the current portfolio by default
        return self.portfolio._build_dto(self.last_effective_date)

    def reset_trades(self, target_portfolio: PortfolioDTO | None = None):
        """
        Will delete all existing trades and recreate them from the method `create_or_update_trades`
        """
        if self.status != TradeProposal.Status.DRAFT:
            raise ValueError("Cannot reset non-draft trade proposal. Revert this trade proposal first.")
        # delete all existing trades
        self.trades.all().delete()
        last_effective_date = self.last_effective_date
        # Get effective and target portfolio
        effective_portfolio = self.portfolio._build_dto(last_effective_date)
        if not target_portfolio:
            target_portfolio = self._get_target_portfolio()
        # if not effective_portfolio:
        #     effective_portfolio = target_portfolio
        service = TradingService(
            self.trade_date,
            effective_portfolio=effective_portfolio,
            target_portfolio=target_portfolio,
        )
        service.normalize()
        service.is_valid()

        for trade_dto in service.validated_trades:
            instrument = Instrument.objects.get(id=trade_dto.underlying_instrument)
            currency_fx_rate = instrument.currency.convert(
                last_effective_date, self.portfolio.currency, exact_lookup=True
            )
            trade = Trade(
                underlying_instrument=instrument,
                transaction_subtype=Trade.Type.BUY if trade_dto.delta_weight > 0 else Trade.Type.SELL,
                currency=instrument.currency,
                value_date=last_effective_date,
                transaction_date=self.trade_date,
                trade_proposal=self,
                portfolio=self.portfolio,
                weighting=trade_dto.delta_weight,
                status=Trade.Status.DRAFT,
                currency_fx_rate=currency_fx_rate,
            )
            trade.shares = self.estimate_shares(trade)
            trade.save()

    def replay(self):
        last_trade_proposal = self
        while last_trade_proposal and last_trade_proposal.status == TradeProposal.Status.APPROVED:
            logger.info(f"Replaying trade proposal {last_trade_proposal}")
            last_trade_proposal.portfolio.assets.filter(
                date=last_trade_proposal.trade_date
            ).delete()  # we delete the existing position and we reapply the trade proposal
            if not last_trade_proposal.portfolio.assets.filter(date=last_trade_proposal.trade_date).exists():
                if last_trade_proposal.status == TradeProposal.Status.APPROVED:
                    logger.info("Reverting trade proposal ...")
                    last_trade_proposal.revert()
                if last_trade_proposal.status == TradeProposal.Status.DRAFT:
                    logger.info("Submitting trade proposal ...")
                    last_trade_proposal.submit()
                if last_trade_proposal.status == TradeProposal.Status.SUBMIT:
                    logger.info("Approving trade proposal ...")
                    last_trade_proposal.approve()
                last_trade_proposal.save()
            next_trade_proposal = last_trade_proposal.next_trade_proposal
            next_trade_date = (
                next_trade_proposal.trade_date - timedelta(days=1) if next_trade_proposal else date.today()
            )
            overriding_trade_proposal = last_trade_proposal.portfolio.batch_portfolio(
                last_trade_proposal.trade_date, next_trade_date
            )
            last_trade_proposal = overriding_trade_proposal or next_trade_proposal

    def estimate_shares(self, trade: Trade) -> Decimal | None:
        if not self.portfolio.only_weighting and (quote := trade.underlying_quote_price):
            trade_total_value_fx_portfolio = (
                self.portfolio.get_total_asset_value(trade.value_date) * trade._target_weight
            )
            price_fx_portfolio = quote.net_value * trade.currency_fx_rate
            if price_fx_portfolio:
                return trade_total_value_fx_portfolio / price_fx_portfolio

    # Start FSM logics

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.SUBMIT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        self.trades.update(comment="", status=Trade.Status.DRAFT)
        for trade in self.trades.all():
            trade.submit()
            trade.save()
        self.evaluate_active_rules(
            self.trade_date, self.validated_trading_service.target_portfolio, asynchronously=True
        )

    def can_submit(self):
        errors = dict()
        errors_list = []
        if self.trades.exists() and self.trades.exclude(status=Trade.Status.DRAFT).exists():
            errors_list.append("All trades need to be draft before submitting")
        service = self.validated_trading_service
        try:
            service.is_valid(ignore_error=True)
            # if service.trades_batch.totat_abs_delta_weight == 0:
            #     errors_list.append(
            #         "There is no change detected in this trade proposal. Please submit at last one valid trade"
            #     )
            if len(service.validated_trades) == 0:
                errors_list.append("There is no valid trade on this proposal")
            if service.errors:
                errors_list.extend(service.errors)
            if errors_list:
                errors["non_field_errors"] = errors_list
        except ValidationError:
            errors["non_field_errors"] = service.errors
            with suppress(KeyError):
                del self.__dict__["validated_trading_service"]
        return errors

    @property
    def can_be_approved_or_denied(self):
        return self.has_no_rule_or_all_checked_succeed and self.portfolio.is_manageable

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def approve(self, by=None, description=None, synchronous=False, **kwargs):
        # We validate trade which will create or update the initial asset positions
        if not self.portfolio.can_be_rebalanced:
            raise ValueError("Non-Rebalanceable portfolio cannot be traded manually.")
        self.trades.update(status=Trade.Status.SUBMIT)
        self.portfolio.assets.filter(date=self.trade_date).delete()  # we delete position to avoid having leftovers
        for trade in self.trades.all():
            trade.execute()
            trade.save()
        self.portfolio.change_at_date(self.trade_date)
        # replay_as_task.delay(self.id)

    def can_approve(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["non_field_errors"] = "The portfolio does not allow manual rebalanced"
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = "At least one trade needs to be submitted to be able to approve this proposal"
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = (
                "The portfolio needs to be a model portfolio in order to approve this trade proposal manually"
            )
        if self.has_assigned_active_rules and not self.has_all_check_completed_and_succeed:
            errors["non_field_errors"] = "The pre trades rules did not passed successfully"
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        self.trades.all().delete()
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]

    def can_deny(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = "At least one trade needs to be submitted to be able to deny this proposal"
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.has_all_check_completed,  # we wait for all checks to succeed before proposing the back to draft transition
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        self.trades.update(status=Trade.Status.DRAFT)
        self.checks.delete()

    def can_backtodraft(self):
        pass

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.REGENERATE.icon,
                key="revert",
                label="Revert",
                action_label="revert",
                description_fields="<p>Unapply trades and move everything back to draft (i.e. The underlying asset positions will change like the trades were never applied)</p>",
            )
        },
    )
    def revert(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        for trade in self.trades.filter(status=Trade.Status.EXECUTED):
            trade.revert()
            trade.save()
        # replay_as_task.delay(self.id)

    def can_revert(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = (
                "The portfolio needs to be a model portfolio in order to revert this trade proposal manually"
            )
        return errors

    # End FSM logics

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:tradeproposal"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:tradeproposalrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{_portfolio.name}} ({{trade_date}})"


@shared_task(queue="portfolio")
def replay_as_task(trade_proposal_id):
    trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
    trade_proposal.replay()



from datetime import date
from decimal import Decimal

from django.core.exceptions import ValidationError

from wbportfolio.pms.typing import Portfolio, Trade, TradeBatch


class TradingService:
    """
    This class represents the trading service. It can be instantiated either with the target portfolio and the effective portfolio or given a direct list of trade
    In any case, it will compute all three states
    """

    def __init__(
        self,
        trade_date: date,
        effective_portfolio: Portfolio | None = None,
        target_portfolio: Portfolio | None = None,
        trades_batch: TradeBatch | None = None,
        total_value: Decimal = None,
    ):
        self.total_value = total_value
        self.trade_date = trade_date
        if target_portfolio is None:
            target_portfolio = Portfolio(positions=())
        if effective_portfolio is None:
            effective_portfolio = Portfolio(positions=())
        # If effective portfoolio and trades batch is provided, we ensure the trade batch contains at least one trade for every position
        trades_batch = self.build_trade_batch(effective_portfolio, target_portfolio, trades_batch=trades_batch)
        # if no trade but a effective portfolio is provided, we get the trade batch only from the effective portofolio (and the target portfolio if provided, but optional. Without it, the trade delta weight will be 0 )
        # Finally, we compute the target portfolio
        if trades_batch and not target_portfolio:
            target_portfolio = trades_batch.convert_to_portfolio()

        self.trades_batch = trades_batch
        self.effective_portfolio = effective_portfolio
        self.target_portfolio = target_portfolio

    @property
    def errors(self) -> list[str]:
        """
        Returned the list of errors stored during the validation process. Can only be called after is_valid
        """
        if not hasattr(self, "_errors"):
            msg = "You must call `.is_valid()` before accessing `.errors`."
            raise AssertionError(msg)
        return self._errors

    @property
    def validated_trades(self) -> list[Trade]:
        """
        Returned the list of validated trade stored during the validation process. Can only be called after is_valid
        """
        if not hasattr(self, "_validated_trades"):
            msg = "You must call `.is_valid()` before accessing `.validated_trades`."
            raise AssertionError(msg)
        return self._validated_trades

    def run_validation(self, validated_trades: list[Trade]):
        """
        Test the given value against all the validators on the field,
        and either raise a `ValidationError` or simply return.
        """
        TradeBatch(validated_trades).validate()
        if self.effective_portfolio:
            for trade in validated_trades:
                if (
                    trade.effective_weight
                    and trade.underlying_instrument not in self.effective_portfolio.positions_map
                ):
                    raise ValidationError("All effective position needs to be matched with a validated trade")

    def build_trade_batch(
        self,
        effective_portfolio: Portfolio,
        target_portfolio: Portfolio,
        trades_batch: TradeBatch | None = None,
    ) -> TradeBatch:
        """
        Given combination of effective portfolio and either a trades batch or a target portfolio, ensure all theres variables are set

        Args:
            effective_portfolio: The effective portfolio
            target_portfolio: The optional target portfolio
            trades_batch: The optional trades batch

        Returns: The normalized trades batch
        """
        instruments = list(effective_portfolio.positions_map.keys())
        instruments.extend(list(target_portfolio.positions_map.keys()))
        if trades_batch:
            instruments.extend(list(trades_batch.trades_map.keys()))
        _trades: list[Trade] = []
        for instrument in set(instruments):
            effective_weight = target_weight = 0
            effective_shares = 0
            instrument_type = currency = None
            if effective_pos := effective_portfolio.positions_map.get(instrument, None):
                effective_weight = target_weight = effective_pos.weighting
                effective_shares = effective_pos.shares
                instrument_type, currency = effective_pos.instrument_type, effective_pos.currency
            if target_pos := target_portfolio.positions_map.get(instrument, None):
                target_weight = target_pos.weighting
                instrument_type, currency = target_pos.instrument_type, target_pos.currency
            if trades_batch and (trade := trades_batch.trades_map.get(instrument, None)):
                effective_weight, target_weight = trade.effective_weight, trade.target_weight
                effective_shares = trade.effective_shares
                instrument_type, currency = trade.instrument_type, trade.currency

            _trades.append(
                Trade(
                    underlying_instrument=instrument,
                    effective_weight=effective_weight,
                    target_weight=target_weight,
                    effective_shares=effective_shares,
                    date=self.trade_date,
                    instrument_type=instrument_type,
                    currency=currency,
                )
            )
        return TradeBatch(tuple(_trades))

    def is_valid(self, ignore_error: bool = False) -> bool:
        """
        Validate the trade batch against a set of default rules. Populate the validated_trades and errors property.
        Ignore error by default
        Args:
            ignore_error: If true, will raise the error. False by default

        Returns: True if the trades batch is valid
        """
        if not hasattr(self, "_validated_trades"):
            self._validated_trades = []
            self._errors = []
            # Run validation for every trade. If a trade is not valid, we simply exclude it from the validated trades list
            for _, trade in self.trades_batch.trades_map.items():
                try:
                    trade.validate()
                    self._validated_trades.append(trade)
                except ValidationError as exc:
                    self._errors.append(exc.message)
            try:
                # Check the overall validity of the trade batch. If this fail, we consider all trade invalids
                self.run_validation(self._validated_trades)
            except ValidationError as exc:
                self._validated_trades = []
                self._errors.append(exc.message)

            if self._errors and not ignore_error:
                raise ValidationError(self.errors)

        return not bool(self._errors)

    def normalize(self):
        """
        Normalize the instantiate trades batch so that the target weight is 100%
        """
        self.trades_batch = TradeBatch(
            [trade.normalize_target(self.trades_batch.total_target_weight) for trade in self.trades_batch.trades]
        )



from wbcore.metadata.configs.titles import TitleViewConfig

from wbportfolio.models import Portfolio


class TransactionPortfolioTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Transactions for Product {str(portfolio)}"



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.display import Display
from wbcore.metadata.configs.display.instance_display.layouts.layouts import Layout
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.pages import Page
from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import repeat
from wbfdm.viewsets.configs.display.instruments import InstrumentDisplayConfig


class ProductGroupDisplayConfig(InstrumentDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="identifier", label="Identifier"),
                dp.Field(key="name", label="Name"),
                dp.Field(key="umbrella", label="Umbrella"),
                dp.Field(key="type", label="Type"),
                dp.Field(key="category", label="Category"),
                dp.Field(key="management_company", label="Management"),
                dp.Field(key="depositary", label="Depositary"),
                dp.Field(key="transfer_agent", label="TA"),
                dp.Field(key="administrator", label="Administrator"),
                dp.Field(key="investment_manager", label="Manager"),
                dp.Field(key="auditor", label="Auditor"),
                dp.Field(key="paying_agent", label="Paying Agent"),
            ]
        )

    def get_instance_display(self) -> Display:
        display = super().get_instance_display()
        return Display(
            pages=[
                display.pages[0],
                Page(
                    title="Product Group Information",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["type", "category", "umbrella", "."],
                                ["management_company", "depositary", "transfer_agent", "administrator"],
                                ["investment_manager", "auditor", "paying_agent", "."],
                            ],
                            grid_template_columns=[
                                repeat(4, Style.fr(1)),
                            ],
                        )
                    },
                ),
                *display.pages[1:],
            ]
        )



from django.contrib.messages import warning
from django.core.exceptions import ValidationError
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.serializers import DefaultFromView

from wbportfolio.models import Portfolio, RebalancingModel, TradeProposal

from .. import PortfolioRepresentationSerializer, RebalancingModelRepresentationSerializer


class TradeProposalModelSerializer(wb_serializers.ModelSerializer):
    rebalancing_model = wb_serializers.PrimaryKeyRelatedField(
        queryset=RebalancingModel.objects.all(),
        default=DefaultFromView("portfolio.automatic_rebalancer.rebalancing_model"),
    )
    _rebalancing_model = RebalancingModelRepresentationSerializer(source="rebalancing_model")
    target_portfolio = wb_serializers.PrimaryKeyRelatedField(
        queryset=Portfolio.objects.all(), write_only=True, required=False, default=DefaultFromView("portfolio")
    )
    _target_portfolio = PortfolioRepresentationSerializer(source="target_portfolio")

    trade_date = wb_serializers.DateField(
        read_only=lambda view: not view.new_mode, default=DefaultFromView("default_trade_date")
    )

    def create(self, validated_data):
        target_portfolio = validated_data.pop("target_portfolio", None)
        rebalancing_model = validated_data.get("rebalancing_model", None)
        if request := self.context.get("request"):
            validated_data["creator"] = request.user.profile
        obj = super().create(validated_data)

        target_portfolio_dto = None
        if (
            target_portfolio
            and not rebalancing_model
            and (
                last_effective_date := target_portfolio.get_latest_asset_position_date(
                    obj.trade_date, with_estimated=True
                )
            )
        ):
            target_portfolio_dto = target_portfolio._build_dto(last_effective_date)
        try:
            obj.reset_trades(target_portfolio=target_portfolio_dto)
        except ValidationError as e:
            if request := self.context.get("request"):
                warning(request, str(e))
        return obj

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        res = {}
        if instance.status == TradeProposal.Status.APPROVED:
            res["replay"] = reverse("wbportfolio:tradeproposal-replay", args=[instance.id], request=request)
        if instance.status == TradeProposal.Status.DRAFT:
            res["reset"] = reverse("wbportfolio:tradeproposal-reset", args=[instance.id], request=request)
        res["trades"] = reverse(
            "wbportfolio:tradeproposal-trade-list",
            args=[instance.id],
            request=request,
        )
        return res

    class Meta:
        model = TradeProposal
        only_fsm_transition_on_instance = True
        fields = (
            "id",
            "trade_date",
            "comment",
            "status",
            "portfolio",
            "_rebalancing_model",
            "rebalancing_model",
            "target_portfolio",
            "_target_portfolio",
            "_additional_resources",
        )


class ReadOnlyTradeProposalModelSerializer(TradeProposalModelSerializer):
    class Meta(TradeProposalModelSerializer.Meta):
        read_only_fields = TradeProposalModelSerializer.Meta.fields



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbfdm.models import Instrument

from wbportfolio.models import TradeProposal


class TradeButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {bt.WidgetButton(key="claims", label="Claims", icon=WBIcon.TRADE.icon)}

    def get_custom_instance_buttons(self):
        btn = self.get_custom_list_instance_buttons()
        btn.add(
            bt.HyperlinkButton(
                key="import_source",
                label="Import Source",
                icon=WBIcon.SAVE.icon,
            )
        )
        return btn


class TradeInstrumentButtonConfig(TradeButtonConfig):
    def get_custom_buttons(self):
        res = set()
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            instrument = Instrument.objects.get(id=instrument_id)
            res = {
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:instrument-custodiandistribution-list", args=[instrument_id], request=self.request
                    ),
                    label="Custodian Distribution",
                ),
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:instrument-customerdistribution-list", args=[instrument_id], request=self.request
                    ),
                    label="Customer Distribution",
                ),
            }
            if instrument.security_instrument_type.key == "product":
                res.add(
                    bt.WidgetButton(
                        endpoint=reverse(
                            "wbportfolio:product-nominalchart-list", args=[instrument_id], request=self.request
                        ),
                        label="Nominal Chart",
                    )
                )
                res.add(
                    bt.WidgetButton(
                        endpoint=reverse(
                            "wbportfolio:product-aumchart-list", args=[instrument_id], request=self.request
                        ),
                        label="AUM Chart",
                    )
                )
        return res


class TradeTradeProposalButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if trade_proposal_id := self.view.kwargs.get("trade_proposal_id", None):
            trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
            if trade_proposal.status == TradeProposal.Status.DRAFT:
                return {
                    bt.DropDownButton(
                        label="Tools",
                        buttons=(
                            bt.ActionButton(
                                method=RequestType.PATCH,
                                identifiers=("wbportfolio:tradeproposal",),
                                endpoint=reverse(
                                    "wbportfolio:tradeproposal-reset", args=[trade_proposal_id], request=self.request
                                ),
                                icon=WBIcon.REGENERATE.icon,
                                label="Reset Trade",
                                description_fields="""
                                        <p>Delete and recreate initial trades to from its associated model portfolio</p>
                                        """,
                                action_label="Reset Trade",
                                title="Reset Trade",
                            ),
                            bt.ActionButton(
                                method=RequestType.PATCH,
                                identifiers=("wbportfolio:tradeproposal",),
                                endpoint=reverse(
                                    "wbportfolio:tradeproposal-normalize",
                                    args=[trade_proposal_id],
                                    request=self.request,
                                ),
                                icon=WBIcon.EDIT.icon,
                                label="Normalize Trades",
                                description_fields="""
                                        <p>Make sure all trades normalize to a total target weight of 100%</p>
                                        """,
                                action_label="Normalize Trades",
                                title="Normalize Trades",
                            ),
                            bt.ActionButton(
                                method=RequestType.PATCH,
                                identifiers=("wbportfolio:tradeproposal",),
                                endpoint=reverse(
                                    "wbportfolio:tradeproposal-deleteall",
                                    args=[trade_proposal_id],
                                    request=self.request,
                                ),
                                icon=WBIcon.DELETE.icon,
                                label="Delete All Trades",
                                description_fields="""
                                <p>Delete all trades from this trade proposal?</p>
                                """,
                                action_label="Delete All Trades",
                                title="Delete All Trades",
                            ),
                        ),
                    )
                }
        return {}


class TradeProposalButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                key="replay",
                icon=WBIcon.SYNCHRONIZE.icon,
                label="Replay Trades",
                description_fields="""
                <p>Replay Trades. It will recompute all assets positions until next trade proposal day (or today otherwise) </p>
                """,
                action_label="Replay Trade",
                title="Replay Trade",
            )
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from wbcore import filters
from wbcore.contrib.directory.filters import CompanyFilter as CF
from wbcore.contrib.directory.filters import PersonFilter as PF
from wbcore.contrib.directory.models import Company, Person


class EntryPortfolioFilter(filters.FilterSet):
    asset_under_management_gte = filters.NumberFilter(
        label="AUM",
        help_text="Filter for the Assets under Management",
        field_name="asset_under_management",
        lookup_expr="gte",
        method="aum_filter_gte",
    )
    asset_under_management_lte = filters.NumberFilter(
        label="AUM",
        help_text="Filter for the Assets under Management",
        field_name="asset_under_management",
        lookup_expr="lte",
        method="aum_filter_lte",
    )

    def aum_filter_gte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__asset_under_management__gte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__asset_under_management__gte=value)
        return queryset

    def aum_filter_lte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__asset_under_management__lte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__asset_under_management__lte=value)
        return queryset

    invested_assets_under_management_usd_gte = filters.NumberFilter(
        label="Invested AUM",
        help_text="Filter for the invested Capital (AUM) with us",
        field_name="invested_assets_under_management_usd",
        lookup_expr="gte",
        method="filter_gte",
    )
    invested_assets_under_management_usd_lte = filters.NumberFilter(
        label="Invested AUM",
        help_text="Filter for the invested Capital (AUM) with us",
        field_name="invested_assets_under_management_usd",
        lookup_expr="lte",
        method="filter_lte",
    )

    def filter_gte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__invested_assets_under_management_usd__gte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__invested_assets_under_management_usd__gte=value)
        return queryset

    def filter_lte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__invested_assets_under_management_usd__lte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__invested_assets_under_management_usd__lte=value)
        return queryset

    potential_gte = filters.NumberFilter(
        label="Potential",
        help_text="Filter for the potential, e.g. how much potential is there to invest with us",
        field_name="potential",
        lookup_expr="gte",
        method="potential_filter_gte",
    )
    potential_lte = filters.NumberFilter(
        label="Potential",
        help_text="Filter for the potential, e.g. how much potential is there to invest with us",
        field_name="potential",
        lookup_expr="lte",
        method="potential_filter_lte",
    )

    def potential_filter_gte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__potential__gte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__potential__gte=value)
        return queryset

    def potential_filter_lte(self, queryset, name, value):
        if value:
            if self.view.get_model() is Person:
                return queryset.filter(employers__portfolio_data__potential__lte=value)
            elif self.view.get_model() is Company:
                return queryset.filter(portfolio_data__potential__lte=value)
        return queryset


class CompanyFilter(CF, EntryPortfolioFilter):
    @classmethod
    def get_filter_class_for_remote_filter(cls):
        """
        Define which filterset class sender to user for remote filter registration
        """
        return CF

    class Meta(CF.Meta):
        fields = {
            **CF.Meta.fields,
        }


class PersonFilter(PF, EntryPortfolioFilter):
    @classmethod
    def get_filter_class_for_remote_filter(cls):
        """
        Define which filterset class sender to user for remote filter registration
        """
        return PF

    class Meta(PF.Meta):
        fields = {
            **PF.Meta.fields,
        }



from datetime import date

from celery import shared_task
from tqdm import tqdm
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Company, Entry

from .models import Updater


@shared_task(queue="portfolio")
def update_all_portfolio_data(val_date: date | None = None):
    if not val_date:
        val_date = CurrencyFXRates.objects.latest("date").date
    updater = Updater(val_date)
    account_owners = Entry.objects.filter(accounts__isnull=False)
    qs = Company.objects.filter(id__in=account_owners)
    for company in tqdm(qs, total=qs.count()):
        updater.update_company_data(company)



from typing import Optional

from django.utils.translation import gettext_lazy as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Layout, Page, default
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_page_with_inline,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class PortfolioDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="currency", label="Currency"),
                dp.Field(key="updated_at", label="Updated At"),
                dp.Field(key="depends_on", label="Depends on"),
                dp.Field(key="automatic_rebalancer", label="Rebalancer"),
                dp.Field(key="invested_timespan", label="Invested"),
                dp.Field(key="is_manageable", label="Managed"),
                dp.Field(key="is_tracked", label="Tracked"),
                dp.Field(key="only_weighting", label="Only-Weight"),
                dp.Field(key="is_lookthrough", label="Lookthrough"),
                dp.Field(key="is_composition", label="Composition"),
            ]
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    title="Main Information",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "name", "currency", "is_manageable"],
                                ["invested_timespan", "invested_timespan", "create_index", "."]
                                if self.new_mode
                                else [
                                    "invested_timespan",
                                    "invested_timespan",
                                    "automatic_rebalancer",
                                    "automatic_rebalancer",
                                ],
                                ["is_tracked", "is_lookthrough", "only_weighting", "is_composition"],
                                [repeat_field(4, "instruments_section")],
                                [repeat_field(4, "dependencyportfolios_section")],
                                [repeat_field(4, "preferredclassification_section")],
                            ],
                            # grid_template_rows=["min-content"] * 3 + ["1fr", "446px"],
                            # grid_template_columns=[repeat(4, "183px"), "1fr", "1fr"],
                            sections=[
                                create_simple_section(
                                    "instruments_section",
                                    _("Linked Instruments"),
                                    [["instruments"]],
                                    "instruments",
                                    collapsed=True,
                                ),
                                create_simple_section(
                                    "dependencyportfolios_section",
                                    _("Dependency Portfolios"),
                                    [["dependencyportfolios"]],
                                    "dependencyportfolios",
                                    collapsed=True,
                                ),
                                create_simple_section(
                                    "preferredclassification_section",
                                    _("Preferred Classification"),
                                    [["preferredclassification"]],
                                    "preferredclassification",
                                    collapsed=True,
                                ),
                            ],
                        ),
                    },
                ),
                Page(
                    title="Trade",
                    layouts={
                        default(): dp.Layout(
                            grid_template_areas=[
                                ["trade_proposals"],
                            ],
                            grid_template_rows=["auto"],
                            inlines=[
                                dp.Inline(key="trade_proposals", endpoint="trade_proposals"),
                            ],
                        )
                    },
                ),
                Page(
                    title="Composition",
                    display=Display(
                        navigation_type=dp.NavigationType.PANEL,
                        pages=[
                            create_simple_page_with_inline("Assets", "assets"),
                            create_simple_page_with_inline("Top-Down Composition", "topdowncomposition"),
                            create_simple_page_with_inline("Contributor", "contributor"),
                            create_simple_page_with_inline("Distribution Table", "distribution_table"),
                            create_simple_page_with_inline("Distribution Chart", "distribution_chart"),
                            # Page(
                            #     title="Distribution",
                            #     layouts={
                            #         default(): Layout(
                            #             grid_template_areas=[["distribution_table", "distribution_chart"]],
                            #             grid_template_rows=["auto"],
                            #             grid_template_columns=["2fr", "1fr"],
                            #             inlines=[
                            #                 Inline(
                            #                     key="distribution_table", endpoint="distribution_table", title="Table"
                            #                 ),
                            #                 Inline(
                            #                     key="distribution_chart", endpoint="distribution_chart", title="Chart"
                            #                 ),
                            #             ],
                            #         ),
                            #     },
                            # ),
                            create_simple_page_with_inline(
                                "Portfolio Composition vs. Dependant Portfolios", "modelcomposition"
                            ),
                        ],
                    ),
                ),
            ]
        )


class PortfolioPortfolioThroughModelDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="dependency_portfolio", label="Dependency Portfolio"),
                dp.Field(key="portfolio", label="Dependant Portfolio"),
                dp.Field(key="type", label="Type"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["dependency_portfolio", "type"],
            ]
        )


class TopDownPortfolioCompositionPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        rebalancing_column_label = "Last Rebalancing"
        effective_column_label = "Actual"
        if self.view.last_rebalancing_date:
            rebalancing_column_label += f" ({self.view.last_rebalancing_date:%Y-ok, %m-%d})"
        if self.view.last_effective_date:
            effective_column_label += f" ({self.view.last_effective_date:%Y-%m-%d})"
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument", label="Instrument"),
                dp.Field(key="effective_weights", label=rebalancing_column_label),
                dp.Field(key="rebalancing_weights", label=effective_column_label),
            ],
            tree=True,
            tree_group_pinned="left",
            tree_group_field="instrument",
            tree_group_label="Instrument",
            tree_group_parent_pointer="parent_row_id",
        )



from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.models import Product


def parse(import_source):
    data = list()
    df = pd.DataFrame()
    try:
        df = pd.read_excel(
            BytesIO(import_source.file.read()), engine="openpyxl", sheet_name="VALUATION_CONSOLIDATION"
        ).dropna(how="all")
    except (ValueError, IndexError):
        pass

    if not df.empty:
        xx, yy = np.where(df == "HISTORICAL VALUATION")
        df = df.iloc[xx[0] :, yy[0] :]
        df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")
        df = df.drop(df.index[0]).set_index("HISTORICAL VALUATION")
        df.index = pd.to_datetime(df.index, dayfirst=True)

        for date, valuations in df.to_dict("index").items():
            date = date.to_pydatetime()
            for isin, value in valuations.items():
                product = Product.objects.get(isin=isin)
                if date.weekday() not in [5, 6]:
                    data.append(
                        {
                            "instrument": {"instrument_type": "product", "id": product.id},
                            "date": date.strftime("%Y-%m-%d"),
                            "net_value": value,
                            "calculated": False,
                        }
                    )

    return {"data": data}



from django.utils import timezone
from django.utils.functional import cached_property
from rest_framework.serializers import ValidationError
from wbcore import serializers
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer
from wbcrm.models import Account
from wbcrm.serializers.accounts import TerminalAccountRepresentationSerializer

from wbportfolio.models import Product, Trade
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.serializers.products import (
    ProductCustomerRepresentationSerializer,
    ProductRepresentationSerializer,
)
from wbportfolio.serializers.transactions.trades import (
    TradeClaimRepresentationSerializer,
)


class ClaimAPIModelSerializer(serializers.ModelSerializer):
    product = serializers.PrimaryKeyRelatedField(
        required=False, queryset=Product.objects.all(), default=serializers.DefaultFromGET("product")
    )
    product_repr = serializers.StringRelatedField(source="product")
    claimant_repr = serializers.StringRelatedField(source="claimant")
    isin = serializers.SerializerMethodField()

    def get_isin(self, obj):
        return obj.product.isin

    @cached_property
    def user_profile(self) -> Person | None:
        if request := self.context.get("request"):
            return request.user.profile
        return None

    def validate(self, data):
        if "creator" not in data:
            data["creator"] = self.user_profile

        if "claimant" not in data:
            data["claimant"] = self.user_profile

        if not self.instance and (request := self.context.get("request")):
            if "product" not in data and "isin" not in request.data:
                raise ValidationError({"product": "Either product or ISIN has to be supplied"})

            if "isin" in request.data:
                try:
                    data["product"] = Product.objects.get(isin=request.data["isin"])
                except Product.DoesNotExist:
                    raise ValidationError({"isin": "A product with this ISIN does not exist."})
        if trade := data.get("trade", None):
            if not trade.is_claimable:
                raise ValidationError({"trade": "Only a claimable trade can be selected"})
        return data

    class Meta:
        model = Claim
        fields = (
            "id",
            "status",
            "product",
            "product_repr",
            "claimant",
            "claimant_repr",
            "date",
            "bank",
            "reference",
            "shares",
            "isin",
            "external_id",
        )


class ClaimRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbportfolio:claim-detail")

    class Meta:
        model = Claim
        fields = (
            "id",
            "shares",
            "bank",
            "account",
            "_detail",
        )


class ClaimModelSerializer(serializers.ModelSerializer):
    account = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Account.objects.all(),
        default=serializers.DefaultFromGetOrKwargs("account", "account_id"),
        many=False,
    )
    _account = TerminalAccountRepresentationSerializer(source="account")
    _trade = TradeClaimRepresentationSerializer(
        source="trade",
        optional_get_parameters={"product": "underlying_instrument", "date": "pivot_date"},
        depends_on=[{"field": "product", "options": {}}],
    )
    _product = ProductRepresentationSerializer(source="product")
    _creator = EntryRepresentationSerializer(source="creator")
    claimant = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Person.objects.all(),
        default=serializers.CurrentUserDefault("profile"),
        many=False,
        read_only=lambda view: not view.request.user.profile.is_internal,
    )
    _claimant = EntryRepresentationSerializer(source="claimant")

    product = serializers.PrimaryKeyRelatedField(
        required=False, queryset=Product.objects.all(), default=serializers.DefaultFromGET("product"), many=False
    )

    creator = serializers.PrimaryKeyRelatedField(
        required=False, queryset=Person.objects.all(), default=serializers.CurrentUserDefault("profile"), many=False
    )

    currency = serializers.CharField(read_only=True, required=False)
    last_nav = serializers.FloatField(
        default=0, read_only=True, label="NAV", required=False, help_text="The latest product's NAV"
    )
    total_value = serializers.FloatField(
        default=0,
        read_only=True,
        label="Value",
        required=False,
        help_text="The total value (shares * Nav) in the trade currency",
    )
    total_value_usd = serializers.FloatField(
        default=0, read_only=True, label="Value ($)", required=False, help_text="The total value (shares * Nav) in USD"
    )
    date = serializers.DateField(default=lambda: timezone.now().date())
    trade_comment = serializers.TextField(read_only=True, required=False)

    def validate(self, data):
        # If shares or nominal_amount is supplied we can just continue
        if "shares" in data or "nominal_amount" in data:
            # Pop the trade_type and quantity fields, as they are not needed anymore
            data.pop("trade_type", None)
            data.pop("quantity", None)
            return data

        # If we do not have a trade type and instance is None, then we have to raise an error
        if "trade_type" not in data and self.instance is None:
            raise ValidationError({"trade_type": "Trade type is required"})

        trade_type = data.pop("trade_type", self.instance.shares > 0 if self.instance else True)
        as_shares = data.get("as_shares", self.instance.as_shares if self.instance else True)

        # If a new quantity is provided, we take that one, if not we take the value from the instance, depending on the as_shares flag from before this instance was saved
        quantity = data.pop(
            "quantity",
            (
                abs(self.instance.shares)
                if self.instance and self.instance.as_shares
                else abs(self.instance.nominal_amount)
                if self.instance
                else None
            ),
        )
        if quantity is None:
            raise ValidationError({"quantity": "You must provide a quantity"})

        # We set shares/nominal_amount properly based on as_shares and the given quantity
        if as_shares:
            data["nominal_amount"] = None
            data["shares"] = quantity if trade_type else quantity * -1
        else:
            data["shares"] = None
            data["nominal_amount"] = quantity if trade_type else quantity * -1
        return data

    def create(self, validated_data):
        instance = super().create(validated_data)

        # We need to attach the two virtual fields back to the instance, as it won't be reloaded from the db
        instance.trade_type = instance.shares > 0
        instance.quantity = instance.shares if instance.as_shares else instance.nominal_amount
        return instance

    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)

        # We need to attach the two virtual fields back to the instance, as it won't be reloaded from the db
        instance.trade_type = instance.shares > 0
        instance.quantity = instance.shares if instance.as_shares else instance.nominal_amount
        return instance

    trade_type = serializers.BooleanField(
        labels=["Subscription", "Redemption"],
        values=[True, False],
        active_background_color=["#2e7d32", "white", "#d32f2f"],
        active_color=["white", "black", "white"],
        required=False,
    )
    as_shares = serializers.BooleanField(labels=["Shares", "Nominal"], values=[True, False], label="Quantity Type")
    quantity = serializers.DecimalField(
        max_digits=15,
        decimal_places=4,  # the number of decimal for quantity needs to match the number for the shares field
        signed=False,
        depends_on=[{"field": "trade_type", "options": {}}, {"field": "as_shares", "options": {}}],
        required=False,
    )

    class Meta:
        model = Claim
        decorators = {
            "last_nav": serializers.decorator(decorator_type="text", position="left", value="{{currency}}"),
            "total_value": serializers.decorator(decorator_type="text", position="left", value="{{currency}}"),
            "total_value_usd": serializers.decorator(decorator_type="text", position="left", value="$"),
        }
        fields = (
            "id",
            "status",
            "account",
            "_account",
            "trade",
            "_trade",
            "product",
            "_product",
            "claimant",
            "_claimant",
            "creator",
            "_creator",
            "date",
            "bank",
            "reference",
            "shares",
            "last_nav",
            "total_value",
            "total_value_usd",
            "currency",
            "trade_type",
            "quantity",
            "as_shares",
            "nominal_amount",
            "trade_comment",
            "_additional_resources",
            "_buttons",
        )


class ClaimCustomerModelSerializer(ClaimModelSerializer):
    _product = ProductCustomerRepresentationSerializer(source="product")

    class Meta:
        model = Claim
        fields = (
            "id",
            "status",
            "account",
            "_account",
            "account",
            "product",
            "_product",
            "claimant",
            "_claimant",
            "creator",
            "_creator",
            "date",
            "bank",
            "last_nav",
            "total_value",
            "total_value_usd",
            "reference",
            "shares",
            "nominal_amount",
            "trade_type",
            "quantity",
            "as_shares",
            "_additional_resources",
        )


class ClaimTradeModelSerializer(ClaimModelSerializer):
    bank = serializers.CharField(default=serializers.DefaultFromView("trade.bank"))
    product = serializers.PrimaryKeyRelatedField(
        default=serializers.DefaultFromView("trade.product"),
        many=False,
        queryset=Product.objects.all(),
    )
    date = serializers.DateField(default=serializers.DefaultFromView("trade.transaction_date"))

    def validate(self, data):
        trade = Trade.objects.get(id=self.context["view"].kwargs["trade_id"])
        data["date"] = trade.transaction_date
        data["product"] = trade.product
        data["bank"] = trade.bank
        return super().validate(data)


class ClaimAccountSerializer(ClaimModelSerializer):
    _account = TerminalAccountRepresentationSerializer(source="account")


class NegativeTermimalAccountPerProductModelSerializer(serializers.ModelSerializer):
    id = serializers.PrimaryKeyCharField(read_only=True)
    product_repr = serializers.CharField(read_only=True)
    account_repr = serializers.CharField(read_only=True)
    product_id = serializers.PrimaryKeyField(read_only=True)
    account_id = serializers.PrimaryKeyField(read_only=True)
    sum_shares = serializers.DecimalField(read_only=True, max_digits=12, decimal_places=2)

    class Meta:
        model = Claim
        fields = ("id", "product_repr", "account_repr", "product_id", "account_id", "sum_shares")



from typing import Generator

from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register

from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Valuation as ValuationDTO

from .mixins import ActiveProductRelationshipMixin, StopLossMixin


@register("Stop Loss Portfolio Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActiveProductRelationshipMixin, StopLossMixin, backend.AbstractRuleBackend):
    def _build_dto_args(self) -> tuple[PortfolioDTO, ValuationDTO | None]:
        return (
            self.portfolio._build_dto(self.evaluation_date),
            super()._build_dto_args()[1],
        )

    def is_passive_evaluation_valid(self) -> bool:
        return (
            super().is_passive_evaluation_valid()
            and self.portfolio
            and self.portfolio.assets.filter(date=self.evaluation_date).exists()
        )

    def _process_dto(
        self, portfolio: PortfolioDTO, benchmark_valuation_dto: ValuationDTO = None, **kwargs
    ) -> Generator[backend.IncidentResult, None, None]:
        if self.asset_class:
            portfolio = PortfolioDTO(filter(lambda x: x.instrument_type == self.asset_class.id, portfolio.positions))
        perf_benchmark = self._get_performance(benchmark_valuation_dto)
        for instrument_id, pos in portfolio.positions_map.items():
            if pos.price is not None:
                perf_instrument = self._get_performance(ValuationDTO(instrument=instrument_id, net_value=pos.price))

                yield from self._generate_incidents(instrument_id, perf_instrument, perf_benchmark)



import datetime as dt

from django.db.models import TextChoices
from pandas.tseries.offsets import BDay
from wbcore import filters as wb_filters
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.pandas.filterset import PandasFilterSetMixin
from wbcore.utils.date import current_financial_month

from wbportfolio.filters.assets import (
    DateFilterMixin,
    get_latest_end_quarter_date_asset_position,
)
from wbportfolio.models import AssetPosition


class GroupbyChoice(TextChoices):
    UNDERLYING_INSTRUMENT = "UNDERLYING_INSTRUMENT", "Underlying Instrument"
    CURRENCY = "CURRENCY", "Currency"
    COUNTRY = "COUNTRY", "Country"
    PORTFOLIO = "PORTFOLIO", "Portfolio"
    PRIMARY_CLASSIFICATION = "PRIMARY_CLASSIFICATION", "Primary Classification (Specified Height)"
    PREFERRED_CLASSIFICATION = "PREFERRED_CLASSIFICATION", "Preferred Classification (Specified Height)"

    @classmethod
    def get_id(cls, name: str) -> str:
        return {
            "UNDERLYING_INSTRUMENT": "underlying_instrument",
            "CURRENCY": "underlying_instrument__currency",
            "COUNTRY": "underlying_instrument__country",
            "PORTFOLIO": "portfolio",
            "PRIMARY_CLASSIFICATION": "classification_id",
            "PREFERRED_CLASSIFICATION": "classification_id",
        }[name]

    @classmethod
    def get_repr(cls, name: str) -> str:
        return {
            "UNDERLYING_INSTRUMENT": "underlying_instrument__name_repr",
            "CURRENCY": "underlying_instrument__currency__key",
            "COUNTRY": "underlying_instrument__country__name",
            "PORTFOLIO": "portfolio__name",
            "PRIMARY_CLASSIFICATION": "classification_title",
            "PREFERRED_CLASSIFICATION": "classification_title",
        }[name]


def get_latest_date_based_on_multiple_models(field, request, view):
    models = [AssetPosition.objects, CurrencyFXRates.objects]
    default_date = (dt.date.today() - BDay(1)).date()  # default latest date
    if not any(model.exists() for model in models):  # if at least one model is empty, it return the default date.
        return default_date

    def get_most_recent_common_date(date):
        for model in models:
            latest_model_date = model.filter(date__lte=date).latest("date").date
            if latest_model_date != date:
                return get_most_recent_common_date(latest_model_date)
        return date

    latest_date = get_most_recent_common_date(default_date)
    return latest_date


class AssetPositionPandasFilter(DateFilterMixin, PandasFilterSetMixin, wb_filters.FilterSet):
    # price_start = wb_filters.NumberFilter(label="Price Start")
    # price_end = wb_filters.NumberFilter(label="Price End")

    date = total_value_fx_usd = total_value_fx_usd__gte = total_value_fx_usd__lte = None

    date = wb_filters.FinancialPerformanceDateRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        required=True,
        clearable=False,
        default=current_financial_month,
    )

    group_by = wb_filters.ChoiceFilter(
        label="Group By",
        choices=GroupbyChoice.choices,
        default=GroupbyChoice.UNDERLYING_INSTRUMENT,
        method="fake_filter",
        clearable=False,
        required=True,
    )

    groupby_classification_height = wb_filters.NumberFilter(
        method="fake_filter", label="Classification Height (groupby)", default=0, clearable=False, required=True
    )

    class Meta:
        model = AssetPosition
        fields = {
            # "price_start" : ["gte", "exact", "lte"],
            # "price_end" : ["gte", "exact", "lte"],
            "underlying_instrument": ["exact"],
            "portfolio": ["exact"],
            "currency": ["exact"],
            "exchange": ["exact"],
        }
        df_fields = {
            "total_value_start__gte": wb_filters.NumberFilter(
                precision=1, label="Total Value Start", lookup_expr="gte", field_name="total_value_start"
            ),
            "total_value_end__gte": wb_filters.NumberFilter(
                precision=1, label="Total Value End", lookup_expr="gte", field_name="total_value_end"
            ),
            "allocation_start__gte": wb_filters.NumberFilter(
                precision=2, label="Weighting Start", lookup_expr="gte", field_name="allocation_start"
            ),
            "allocation_end__gte": wb_filters.NumberFilter(
                precision=2, label="Weighting End", lookup_expr="gte", field_name="allocation_end"
            ),
            "performance_total__gte": wb_filters.NumberFilter(
                precision=2, label="Total Performance", lookup_expr="gte", field_name="performance_total"
            ),
            "contribution_total__gte": wb_filters.NumberFilter(
                precision=2, label="Total Contribution", lookup_expr="gte", field_name="contribution_total"
            ),
            "performance_forex__gte": wb_filters.NumberFilter(
                precision=2, label="Forex Performance", lookup_expr="gte", field_name="performance_forex"
            ),
            "contribution_forex__gte": wb_filters.NumberFilter(
                precision=2, label="Forex  Contribution", lookup_expr="gte", field_name="contribution_forex"
            ),
            "market_share__gte": wb_filters.NumberFilter(
                precision=2, label="Market Shares", lookup_expr="gte", field_name="market_share"
            ),
            "total_value_start__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="total_value_start", precision=1, label="Total Value Start"
            ),
            "total_value_end__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="total_value_end", precision=1, label="Total Value End"
            ),
            "allocation_start__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="allocation_start", precision=2, label="Weighting Start"
            ),
            "allocation_end__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="allocation_end", precision=2, label="Weighting End"
            ),
            "performance_total__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="performance_total", precision=2, label="Total Performance"
            ),
            "contribution_total__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="contribution_total", precision=2, label="Total Contribution"
            ),
            "performance_forex__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="performance_forex", precision=2, label="Forex Performance"
            ),
            "contribution_forex__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="contribution_forex", precision=2, label="Forex  Contribution"
            ),
            "market_share__lte": wb_filters.NumberFilter(
                lookup_expr="lte", field_name="market_share", precision=2, label="Market Shares"
            ),
        }


class AggregatedAssetPositionLiquidityFilter(PandasFilterSetMixin, wb_filters.FilterSet):
    historic_date = wb_filters.DateFilter(
        label="Historic Date",
        method=lambda queryset, label, value: queryset,
        default=get_latest_date_based_on_multiple_models,
        required=True,
    )
    compared_date = wb_filters.DateFilter(
        label="Compared Date",
        method=lambda queryset, label, value: queryset,
        default=get_latest_end_quarter_date_asset_position,
        required=True,
    )
    bigger_than_x = wb_filters.NumberFilter(
        label="Bigger Than..", method=lambda queryset, label, value: queryset, default=1.00, required=True, precision=2
    )

    class Meta:
        model = AssetPosition
        fields = {}



from typing import TYPE_CHECKING

import pytest
from django.db.utils import IntegrityError

if TYPE_CHECKING:
    from wbportfolio.factories.portfolio_swing_pricings import (
        PortfolioCashTargetFactory,
    )
    from wbportfolio.models.portfolio import Portfolio
    from wbportfolio.models.portfolio_swing_pricings import PortfolioCashTarget


@pytest.mark.django_db
class TestPortfolioCashTarget:
    def test_factory(self, portfolio_cash_target: "PortfolioCashTarget"):
        assert portfolio_cash_target.pk is not None

    def test_constraint_unique_valid_date_portfolio(
        self, portfolio: "Portfolio", portfolio_cash_target_factory: "PortfolioCashTargetFactory"
    ):
        target1 = portfolio_cash_target_factory.create(portfolio=portfolio)
        target2 = portfolio_cash_target_factory.create(portfolio=portfolio)

        target2.valid_date = target1.valid_date
        with pytest.raises(IntegrityError):
            target2.save()



import datetime
import re

from django.db.models import Q

from wbportfolio.models import Product

INSTRUMENT_MAP_NAME = {"EDA23_AtonRa Z class": Product.objects.get(isin="LU2170995018")}


def _get_exchange_from_ticker(ticker):
    try:
        ticker, exchange = ticker.split(" ")
    except Exception:
        ticker, exchange = ticker, None
    return exchange


def _get_ticker(ticker):
    return ticker.split(" ")[0]


def _get_underlying_instrument(bbg_code, name, currency, instrument_type="equity", isin=None, cash_position=False):
    if product := INSTRUMENT_MAP_NAME.get(bbg_code, None):
        return {"id": product.id}

    exchange = _get_exchange_from_ticker(bbg_code)
    ticker = _get_ticker(bbg_code)

    cash_position = cash_position or "CASH" == bbg_code or len(re.findall("(CASH [A-Z]{3})", bbg_code)) > 0
    if cash_position:
        underlying_quote = {"instrument_type": "cash", "currency__key": currency}
    else:
        underlying_quote = {
            "exchange": {"bbg_exchange_codes": exchange},
            "currency__key": currency,
            "name": name.split(" @")[0].split("/")[
                0
            ],  # we remove anything after @ and also any trailing symbol (which represents some country information at natixis)
            "ticker": ticker,
            "instrument_type": instrument_type,
        }
        if not isin:
            isin_re = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", ticker)
            if len(isin_re) > 0:
                isin = isin_re[0]
                # Natixis gives us ISIN as ticker for product. in that case, we registered the isin but we remove the ticker
                underlying_quote["instrument_type"] = "product"
                del underlying_quote["ticker"]
            elif Product.objects.filter(ticker=ticker).exists():
                underlying_quote["instrument_type"] = "product"
        if isin:
            underlying_quote["isin"] = isin
    return underlying_quote


def file_name_parse(file_name):
    dates = re.findall(r"_([0-9]{4}-?[0-9]{2}-?[0-9]{2})", file_name)
    identifier = re.findall(r"([A-Z]{2}(?![A-Z]{10}\b)[A-Z0-9]{10})_", file_name)
    assert len(dates) == 2, "Not 2 dates found in the filename"
    assert len(identifier) == 1, "Not exactly 1 identifier found in the filename"
    try:
        valuation_date = datetime.datetime.strptime(dates[0], "%Y-%m-%d").date()
    except ValueError:
        valuation_date = datetime.datetime.strptime(dates[0], "%Y%m%d").date()
    try:
        generation_date = datetime.datetime.strptime(dates[1], "%Y-%m-%d").date()
    except ValueError:
        generation_date = datetime.datetime.strptime(dates[1], "%Y%m%d").date()

    product = Product.objects.get(Q(ticker=identifier[0]) | Q(isin=identifier[0]))

    return {
        "product": product,
        "valuation_date": valuation_date,
        "generation_date": generation_date,
    }



import xlrd

from wbportfolio.models import Fees, Product


def get_isin(sheet):
    for row in sheet.get_rows():
        if row[0].value == "ISIN":
            return row[1].value


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    product = Product.objects.get(isin=get_isin(book.sheet_by_name("Certificate")))
    valuation_sheet = book.sheet_by_name("Report")

    date_column = None
    bank_fees_column = None
    management_fees_column = None
    performance_fees_gross_column = None

    data = list()
    for i, row in enumerate(valuation_sheet.get_rows()):
        if i == 0:
            for column in range(0, valuation_sheet.ncols):
                if "Date" in row[column].value:
                    date_column = column
                if "Mngt Fees Natixis" == row[column].value:
                    bank_fees_column = column
                if "Mngt Fees Y" == row[column].value:
                    management_fees_column = column
                if "Performance Fees" in row[column].value:
                    performance_fees_gross_column = column
        else:
            valuation_date = xlrd.xldate.xldate_as_datetime(row[date_column].value, 0)
            bank_fees = row[bank_fees_column].value
            management_fees = row[management_fees_column].value

            base_data = {
                "portfolio": product.primary_portfolio.id,
                "linked_product": product.id,
                "transaction_date": valuation_date.strftime("%Y-%m-%d"),
                "calculated": False,
            }
            if management_fees != 0:
                data.append({"transaction_subtype": Fees.Type.MANAGEMENT, "total_value": management_fees, **base_data})
            if bank_fees != 0:
                data.append({"transaction_subtype": Fees.Type.ISSUER, "total_value": bank_fees, **base_data})
            if performance_fees_gross_column and row[performance_fees_gross_column].value != 0:
                data.append(
                    {
                        "transaction_subtype": Fees.Type.PERFORMANCE,
                        "total_value_gross": row[performance_fees_gross_column].value,
                        **base_data,
                    }
                )

    return {"data": data}



from datetime import date

import numpy as np
import pandas as pd
from django.db.models import Exists, OuterRef, QuerySet
from wbfdm.contrib.metric.backends.base import AbstractBackend, Metric
from wbfdm.contrib.metric.exceptions import MetricInvalidParameterException
from wbfdm.models import InstrumentType

from wbportfolio.models import AssetPosition, InstrumentPortfolioThroughModel, Portfolio


class PortfolioMetricBaseBackend(AbstractBackend[Portfolio]):
    BASKET_MODEL_CLASS = Portfolio

    def convert_df_into_metrics(
        self,
        df: pd.DataFrame | pd.Series,
        key: str,
        portfolio_id: int,
        metric_date: date,
        columns_map: dict[int | str, str] | None = None,
        dependency_metrics: list["Metric"] | None = None,
    ) -> list[Metric]:
        if not dependency_metrics:
            dependency_metrics = []
        metrics = []
        df = df.replace([np.inf, -np.inf, np.nan], None)
        if isinstance(df, pd.DataFrame):
            if columns_map:
                df = df.rename(columns=columns_map)
            for instrument_id, metric in df.to_dict("index").items():
                metrics.append(
                    Metric(
                        basket_id=portfolio_id,
                        basket_content_type_id=self.content_type.id,
                        key=key,
                        metrics=metric,
                        date=metric_date,
                        instrument_id=instrument_id,
                        dependency_metrics=dependency_metrics,
                    )
                )
        else:
            if columns_map:
                df = df.rename(index=columns_map)
            metrics.append(
                Metric(
                    basket_id=portfolio_id,
                    basket_content_type_id=self.content_type.id,
                    key=key,
                    metrics=df.to_dict(),
                    date=metric_date,
                    dependency_metrics=dependency_metrics,
                )
            )
        return metrics

    def _get_valid_date(self, portfolio):
        qs = portfolio.assets.filter(is_estimated=False)
        if self.val_date is not None:
            qs = qs.filter(date__lte=self.val_date)
        try:
            return qs.latest("date").date
        except AssetPosition.DoesNotExist:
            raise MetricInvalidParameterException()

    def get_queryset(self) -> QuerySet[Portfolio]:
        product_portfolios = (
            super()
            .get_queryset()
            .annotate(
                has_product=Exists(
                    InstrumentPortfolioThroughModel.objects.filter(
                        instrument__instrument_type=InstrumentType.PRODUCT, portfolio=OuterRef("pk")
                    )
                )
            )
            .filter(has_product=True, is_active=True)
        )
        try:
            last_position_date = (
                AssetPosition.objects.filter(portfolio__in=product_portfolios, is_estimated=False).latest("date").date
            )
        except AssetPosition.DoesNotExist:
            last_position_date = date.today()
        return product_portfolios.filter_invested_at_date(last_position_date)



import numpy as np
import pandas as pd

from wbportfolio.models import Trade

from .utils import file_name_parse

UNVALID_CUSTODIANS = ["init"]

FIELD_MAP = {
    "Trade Date": "transaction_date",
    "Value Date": "value_date",
    "Counterparty": "bank",
    "Identifier": "external_id",
    "Note price in%": "price",
    "Nominal Increase/Decrease": "shares",
}


def _check_if_count_towards_total_aum(row, df):
    previous_accumulated_shares = row["Accumulated Nominal"] - row["shares"]
    dff = df[
        (df["transaction_date"] < row["transaction_date"]) & (df["Accumulated Nominal"] == previous_accumulated_shares)
    ]
    return not dff["shares"].sum() == row["shares"]


def parse(import_source):
    # Load file into a CSV DictReader
    df = pd.read_csv(import_source.file, encoding="utf-8", delimiter=";")
    df = df.rename(columns=FIELD_MAP)
    parts = file_name_parse(import_source.file.name)

    # Get the valuation date and product from the parts list
    valuation_date = parts["valuation_date"]
    product = parts["product"]

    df["transaction_date"] = pd.to_datetime(df["transaction_date"], format="%d/%m/%Y").dt.strftime("%Y-%m-%d")
    df["value_date"] = pd.to_datetime(df["value_date"], format="%d/%m/%Y").dt.strftime("%Y-%m-%d")
    df["bank"] = df["bank"].str.strip()
    df["shares"] = df["shares"] / product.share_price

    df["bank"] = df["bank"].fillna("<not specified>")

    # Use the accumulated nominal (outstanding shares) to detect internal natixis accounting trade that shouldn't be imported
    # If a trade is marked "INIT" and the previous trade sum equals to that trade shares, we assume these two groups are double accounted and we don't import the INIT trade
    df["Accumulated Nominal"] = df["Accumulated Nominal"] / product.share_price
    df["count_towards_total_aum"] = df.apply(_check_if_count_towards_total_aum, args=[df], axis=1)
    init_rows = df[~df["count_towards_total_aum"] & df["bank"].str.lower().isin(UNVALID_CUSTODIANS)].index
    df = df.drop(init_rows)

    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))

    df["portfolio"] = product.primary_portfolio.id
    df["currency__key"] = product.currency.key
    df["underlying_instrument"] = product.id
    df.loc[df["shares"] < 0, "transaction_subtype"] = Trade.Type.REDEMPTION
    df.loc[df["shares"] >= 0, "transaction_subtype"] = Trade.Type.SUBSCRIPTION

    return {
        "data": df.replace([np.inf, -np.inf, np.nan], None).to_dict(orient="records"),
        "history": {"underlying_instrument": product.id, "transaction_date": valuation_date.strftime("%Y-%m-%d")},
    }
